import time

import pytest
from ophyd.sim import instantiate_fake_device

from haven.devices import lerix

um_per_mm = 1000


def test_rowland_circle_forward():
    rowland = lerix.RowlandPositioner(
        name="rowland", x_motor_pv="", y_motor_pv="", z_motor_pv="", z1_motor_pv=""
    )
    # Check one set of values
    um_per_mm
    result = rowland.forward(500, 60.0, 30.0)
    assert result == pytest.approx(
        (
            500.0 * um_per_mm,  # x
            375.0 * um_per_mm,  # y
            216.50635094610968 * um_per_mm,  # z
            1.5308084989341912e-14 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 80.0, 0.0)
    assert result == pytest.approx(
        (
            484.92315519647707 * um_per_mm,  # x
            0.0 * um_per_mm,  # y
            171.0100716628344 * um_per_mm,  # z
            85.5050358314172 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 70.0, 10.0)
    assert result == pytest.approx(
        (
            484.92315519647707 * um_per_mm,  # x
            109.92315519647711 * um_per_mm,  # y
            291.6982175363274 * um_per_mm,  # z
            75.19186659021767 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 75.0, 15.0)
    assert result == pytest.approx(
        (
            500.0 * um_per_mm,  # x
            124.99999999999994 * um_per_mm,  # y
            216.50635094610965 * um_per_mm,  # z
            2.6514380968122676e-14 * um_per_mm,  # z1
        )
    )
    # Check one set of values
    result = rowland.forward(500, 71.0, 10.0)
    assert result == pytest.approx(
        (
            487.7641290737884 * um_per_mm,  # x
            105.28431301548724 * um_per_mm,  # y
            280.42235703910393 * um_per_mm,  # z
            68.41033299999741 * um_per_mm,  # z1
        )
    )


@pytest.mark.xfail
def test_rowland_circle_inverse():
    rowland = instantiate_fake_device(
        lerix.RowlandPositioner,
        name="rowland",
        x_motor_pv="",
        y_motor_pv="",
        z_motor_pv="",
        z1_motor_pv="",
    )
    # Check one set of values
    result = rowland.inverse(
        x=500.0,  # x
        y=375.0,  # y
        z=216.50635094610968,  # z
        z1=1.5308084989341912e-14,  # z1
    )
    assert result == pytest.approx((500, 60, 30))
    # # Check one set of values
    # result = rowland.forward(500, 80.0, 0.0)
    # assert result == pytest.approx((
    #     484.92315519647707 * um_per_mm,  # x
    #     0.0 * um_per_mm,  # y
    #     171.0100716628344 * um_per_mm,  # z
    #     85.5050358314172 * um_per_mm,  # z1
    # ))
    # # Check one set of values
    # result = rowland.forward(500, 70.0, 10.0)
    # assert result == pytest.approx((
    #     484.92315519647707 * um_per_mm,  # x
    #     109.92315519647711 * um_per_mm,  # y
    #     291.6982175363274 * um_per_mm,  # z
    #     75.19186659021767 * um_per_mm,  # z1
    # ))
    # # Check one set of values
    # result = rowland.forward(500, 75.0, 15.0)
    # assert result == pytest.approx((
    #     500.0 * um_per_mm,  # x
    #     124.99999999999994 * um_per_mm,  # y
    #     216.50635094610965 * um_per_mm,  # z
    #     2.6514380968122676e-14 * um_per_mm,  # z1
    # ))
    # # Check one set of values
    # result = rowland.forward(500, 71.0, 10.0)
    # assert result == pytest.approx((
    #     487.7641290737884 * um_per_mm,  # x
    #     105.28431301548724 * um_per_mm,  # y
    #     280.42235703910393 * um_per_mm,  # z
    #     68.41033299999741 * um_per_mm,  # z1
    # ))


def test_rowland_circle_component():
    device = instantiate_fake_device(
        lerix.LERIXSpectrometer, prefix="255idVME", name="lerix"
    )
    device.rowland.x.user_setpoint._use_limits = False
    device.rowland.y.user_setpoint._use_limits = False
    device.rowland.z.user_setpoint._use_limits = False
    device.rowland.z1.user_setpoint._use_limits = False
    # Set pseudo axes
    statuses = [
        device.rowland.D.set(500.0),
        device.rowland.theta.set(60.0),
        device.rowland.alpha.set(30.0),
    ]
    # [s.wait() for s in statuses]  # <- this should work, need to come back to it
    time.sleep(0.1)
    # Check that the virtual axes were set
    result = device.rowland.get(use_monitor=False)
    assert result.x.user_setpoint == pytest.approx(500.0 * um_per_mm)
    assert result.y.user_setpoint == pytest.approx(375.0 * um_per_mm)
    assert result.z.user_setpoint == pytest.approx(216.50635094610968 * um_per_mm)
    assert result.z1.user_setpoint == pytest.approx(1.5308084989341912e-14 * um_per_mm)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import time
from pathlib import Path

import pytest
from epics import caget, caput

log = logging.getLogger(__name__)


ioc_dir = Path(__file__).parent.resolve() / "iocs"


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_simulated_ioc(ioc_simple):
    assert caget(ioc_simple.pvs["B"], use_monitor=False) == 2.0
    caput(ioc_simple.pvs["B"], 5)
    time.sleep(0.1)
    assert caget(ioc_simple.pvs["B"], use_monitor=False) == 5


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_motor_ioc(ioc_motor):
    prefix = "255idVME:"
    # Check that the starting value is different than what we'll set it to
    assert caput(f"{prefix}m1", 1000)
    assert caget(f"{prefix}m1", use_monitor=False) != pytest.approx(4000)
    # Change the value
    caput(f"{prefix}m1", 4000.0)
    time.sleep(1)
    # Check that the record got updated
    assert caget(f"{prefix}m1.VAL", use_monitor=False) == 4000.0
    assert caget(f"{prefix}m1.RBV", use_monitor=False) == 4000.0


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_scaler_ioc(ioc_scaler):
    # Check that all the channels have the right counts
    for ch_num in range(1, 32):
        pv = f"255idVME:scaler1.S{ch_num}"
        assert caget(pv) is not None, pv


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_mono_ioc(ioc_mono):
    # Test a regular motor
    caput("255idMono:ACS:m1", 0)
    caput("255idMono:ACS:m1.VAL", 0)
    assert caget("255idMono:ACS:m1", use_monitor=False) == 0.0
    assert caget("255idMono:ACS:m1.VAL", use_monitor=False) == 0.0
    # Change the value
    caput("255idMono:ACS:m1", 4000.0)
    time.sleep(0.1)
    # Check that the record got updated
    assert caget("255idMono:ACS:m1", use_monitor=False) == 4000.0
    assert caget("255idMono:ACS:m1.VAL", use_monitor=False) == 4000.0
    assert caget("255idMono:ACS:m1.RBV", use_monitor=False) == 4000.0
    # Test the energy motor
    caput("255idMono:Energy", 10000.0)
    time.sleep(0.1)
    assert caget("255idMono:Energy", use_monitor=False) == 10000.0
    # Change the value
    caput("255idMono:Energy", 6000.0)
    time.sleep(0.1)
    # Check that the record got updated
    assert caget("255idMono:Energy.VAL", use_monitor=False) == 6000.0
    assert caget("255idMono:Energy.RBV", use_monitor=False) == 6000.0


# @pytest.mark.skipif(
#     True or os.environ.get("GITHUB_ACTIONS", "false") == "true",
#     reason="Caproto is too slow on CI.",
# )
# def test_ioc_timing():
#     """Check that the IOC's don't take too long to load."""
#     # Launch the IOC numerous times to see how reproducible it is
#     for pass_num in range(5):
#         start = time.time()
#         with simulated_ioc(fp=ioc_dir / "simple.py"):
#             caput("simple:A", 100)
#             new_value = caget("simple:A", use_monitor=False)
#         assert new_value == 100.0
#         log.info(f"Finish pass {pass_num} in {time.time() - start} seconds.")
#         pass_time = time.time() - start
#         msg = f"Pass {pass_num} took {pass_time} seconds."
#         assert pass_time < 4, msg


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_undulator_ioc(ioc_undulator):
    val = caget(ioc_undulator.pvs["energy"], use_monitor=False)
    assert val == 0.0


# def test_mono_undulator_ioc_again(ioc_undulator):
#     """Check that both mono and undulator IOC's can load in a second set
#     of tests.

#     This is in response to a specific test bug where this would fail.

#     """
#     pass


# def test_mono_undulator_ioc_a_third_time(ioc_undulator):
#     """Check that both mono and undulator IOC's can load in a second set
#     of tests.

#     This is in response to a specific test bug where this would fail.

#     """
#     pass


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
@pytest.mark.xfail
def test_bss_ioc(ioc_bss):
    caput(ioc_bss.pvs["esaf_cycle"], "2023-2", wait=True)
    val = caget(ioc_bss.pvs["esaf_cycle"], as_string=True, use_monitor=False)
    assert val == "2023-2"


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_preamp_ioc(ioc_preamp):
    # Update PVs to recover from other tests
    caput(ioc_preamp.pvs["preamp1_sens_num"], "5")
    caput(ioc_preamp.pvs["preamp1_sens_unit"], "nA/V")
    caput(ioc_preamp.pvs["preamp4_sens_num"], "5")
    caput(ioc_preamp.pvs["preamp4_sens_unit"], "nA/V")
    # Check that the values were set
    assert caget(ioc_preamp.pvs["preamp1_sens_num"], use_monitor=False) == 2
    assert (
        caget(ioc_preamp.pvs["preamp1_sens_num"], use_monitor=False, as_string=True)
        == "5"
    )
    assert (
        caget(ioc_preamp.pvs["preamp4_sens_num"], use_monitor=False, as_string=True)
        == "5"
    )
    assert (
        caget(ioc_preamp.pvs["preamp1_sens_unit"], use_monitor=False, as_string=True)
        == "nA/V"
    )
    assert (
        caget(ioc_preamp.pvs["preamp4_sens_unit"], use_monitor=False, as_string=True)
        == "nA/V"
    )


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_ptc10_ioc(ioc_ptc10):
    assert caput(ioc_ptc10.pvs["tc1_temperature"], 21.3)
    # Check that the values were set
    assert caget(ioc_ptc10.pvs["pid1_voltage"], use_monitor=False) == 0
    assert caget(ioc_ptc10.pvs["pid1_voltage_rbv"], use_monitor=False) == 0
    assert caget(ioc_ptc10.pvs["tc1_temperature"], use_monitor=False) == 21.3


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_area_detector_ioc(ioc_area_detector):
    assert caget(ioc_area_detector.pvs["cam_acquire_busy"], use_monitor=False) == 0


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_dxp_ioc_mca_propogation(ioc_dxp):
    # See if settings propogate to the MCAs
    caput("255idDXP:PresetLive", 1.5)
    caput("255idDXP:PresetReal", 2.5)
    # Check that the values were propogated
    preset_time = caget("255idDXP:mca1.PLTM", use_monitor=False)
    assert preset_time == 1.5
    real_time = caget("255idDXP:mca1.PRTM", use_monitor=False)
    assert real_time == 2.5


@pytest.mark.skip(reason="Simulated IOCs are deprecated.")
def test_dxp_ioc_spectra(ioc_dxp):
    # Get the starting spectrum
    spectrum = caget("255idDXP:mca1.VAL")
    assert not any(spectrum)
    # Start acquring spectra
    caput("255idDXP:PresetReal", 1.0)
    caput("255idDXP:StartAll", 1)
    assert caget("255idDXP:Acquiring", use_monitor=False) == 1
    time.sleep(1.1)
    # Check that acquiring is finished and the spectrum was updated
    assert caget("255idDXP:Acquiring", use_monitor=False) == 0
    spectrum = caget("255idDXP:mca1.VAL")
    assert any(spectrum)



from collections import OrderedDict
from unittest.mock import MagicMock

import numpy as np
import pytest
from ophyd import sim
from ophyd_async.epics.motor import Motor

from haven.plans.fly import FlyerCollector, fly_scan, grid_fly_scan


@pytest.fixture()
def flyer(sim_registry, mocker):
    m1 = Motor("255idcVME:m1", name="m1")
    return m1


def test_set_fly_params(flyer):
    """Does the plan set the parameters of the flyer motor."""
    # step size == 10
    plan = fly_scan([], flyer, -20, 30, num=6, dwell_time=1.5)
    messages = list(plan)
    prep_msg = messages[4]
    assert prep_msg.command == "prepare"
    prep_info = prep_msg.args[0]
    assert prep_info.start_position == -20
    assert prep_info.end_position == 30
    assert prep_info.time_for_move == 9.0


def test_fly_scan_metadata(flyer, ion_chamber):
    """Does the plan set the parameters of the flyer motor."""
    md = {"spam": "eggs"}
    plan = fly_scan([ion_chamber], flyer, -20, 30, num=6, dwell_time=1, md=md)
    messages = list(plan)
    open_msg = messages[1]
    assert open_msg.command == "open_run"
    real_md = open_msg.kwargs
    expected_md = {
        "plan_args": {
            "detectors": list([repr(ion_chamber)]),
            "num": 6,
            "dwell_time": 1,
            "*args": (repr(flyer), -20, 30),
        },
        "plan_name": "fly_scan",
        "motors": [flyer.name],
        "detectors": [ion_chamber.name],
        "spam": "eggs",
    }
    assert real_md == expected_md


def test_collector_describe():
    # Dummy devices with data taken from actual ophyd devices
    aerotech = MagicMock()
    aerotech.describe_collect.return_value = {
        "positions": OrderedDict(
            [
                (
                    "aerotech_horiz",
                    {
                        "source": "PV:25idc:m1.RBV",
                        "dtype": "number",
                        "shape": [],
                        "units": "micron",
                        "lower_ctrl_limit": -28135.352,
                        "upper_ctrl_limit": 31864.648,
                        "precision": 7,
                    },
                ),
                (
                    "aerotech_horiz_user_setpoint",
                    {
                        "source": "PV:25idc:m1.VAL",
                        "dtype": "number",
                        "shape": [],
                        "units": "micron",
                        "lower_ctrl_limit": -28135.352,
                        "upper_ctrl_limit": 31864.648,
                        "precision": 7,
                    },
                ),
            ]
        )
    }
    I0 = MagicMock()
    I0.describe_collect.return_value = {
        "I0": OrderedDict(
            [
                (
                    "I0_net_counts",
                    {
                        "source": "PV:25idcVME:3820:scaler1_netA.D",
                        "dtype": "number",
                        "shape": [],
                        "units": "",
                        "lower_ctrl_limit": 0.0,
                        "upper_ctrl_limit": 0.0,
                        "precision": 0,
                    },
                )
            ]
        )
    }
    motor = MagicMock()
    motor.describe.return_value = OrderedDict(
        [
            (
                "motor",
                {
                    "source": "SIM:motor",
                    "dtype": "integer",
                    "shape": [],
                    "precision": 3,
                },
            ),
            (
                "motor_setpoint",
                {
                    "source": "SIM:motor_setpoint",
                    "dtype": "integer",
                    "shape": [],
                    "precision": 3,
                },
            ),
        ]
    )
    flyers = [aerotech, I0]
    collector = FlyerCollector(
        positioners=[aerotech],
        detectors=[I0],
        stream_name="primary",
        extra_signals=[motor],
        name="collector",
    )
    desc = collector.describe_collect()
    assert "primary" in desc.keys()
    assert list(desc["primary"].keys()) == [
        "aerotech_horiz",
        "aerotech_horiz_user_setpoint",
        "I0_net_counts",
        "motor",
        "motor_setpoint",
    ]
    assert (
        desc["primary"]["aerotech_horiz"]
        == aerotech.describe_collect()["positions"]["aerotech_horiz"]
    )


def test_collector_collect():
    aerotech = MagicMock()
    aerotech.predict.side_effect = [
        # These are the target data, but we need more events to simulate a flying motor
        {
            "data": {
                "aerotech_horiz": -1000.0,
                "aerotech_horiz_user_setpoint": -1000.0,
            },
            "timestamps": {
                "aerotech_horiz": 1691957265.6073308,
                "aerotech_horiz_user_setpoint": 1691957265.6073308,
            },
            "time": 1691957265.6073308,
        },
        {
            "data": {"aerotech_horiz": -800.0, "aerotech_horiz_user_setpoint": -800.0},
            "timestamps": {
                "aerotech_horiz": 1691957266.1137164,
                "aerotech_horiz_user_setpoint": 1691957266.1137164,
            },
            "time": 1691957266.1137164,
        },
    ]
    aerotech.collect.return_value = [
        {
            "data": {
                "aerotech_horiz": -1100.0,
                "aerotech_horiz_user_setpoint": -1100.0,
            },
            "timestamps": {
                "aerotech_horiz": 1691957265.354138,
                "aerotech_horiz_user_setpoint": 1691957265.354138,
            },
            "time": 1691957265.354138,
        },
        {
            "data": {"aerotech_horiz": -900.0, "aerotech_horiz_user_setpoint": -900.0},
            "timestamps": {
                "aerotech_horiz": 1691957265.8605237,
                "aerotech_horiz_user_setpoint": 1691957265.8605237,
            },
            "time": 1691957265.8605237,
        },
        {
            "data": {"aerotech_horiz": -700.0, "aerotech_horiz_user_setpoint": -700.0},
            "timestamps": {
                "aerotech_horiz": 1691957266.366909,
                "aerotech_horiz_user_setpoint": 1691957266.366909,
            },
            "time": 1691957266.366909,
        },
    ]
    I0 = MagicMock()
    I0.collect.return_value = [
        {
            "data": {"I0_net_counts": [0]},
            "timestamps": {"I0_net_counts": [1691957269.1575842]},
            "time": 1691957269.1575842,
        },
        {
            "data": {"I0_net_counts": [0]},
            "timestamps": {"I0_net_counts": [1691957269.0734286]},
            "time": 1691957269.0734286,
        },
    ]
    flyers = [aerotech, I0]
    motor = MagicMock()
    motor.read.return_value = OrderedDict(
        [
            ("motor", {"value": 119.983, "timestamp": 1692072398.879956}),
            ("motor_setpoint", {"value": 120.0, "timestamp": 1692072398.8799553}),
        ]
    )

    collector = FlyerCollector(
        detectors=[I0],
        positioners=[aerotech],
        stream_name="primary",
        name="flyer_collector",
        extra_signals=[motor],
    )
    events = list(collector.collect())
    expected_events = [
        {
            "data": {
                "I0_net_counts": [0],
                "aerotech_horiz": -1000.0,
                "aerotech_horiz_user_setpoint": -1000.0,
                "motor": 119.983,
                "motor_setpoint": 120.0,
            },
            "timestamps": {
                "I0_net_counts": [1691957269.1575842],
                "aerotech_horiz": 1691957265.6073308,
                "aerotech_horiz_user_setpoint": 1691957265.6073308,
                "motor": 1692072398.879956,
                "motor_setpoint": 1692072398.8799553,
            },
            "time": 1691957269.1575842,
        },
        {
            "data": {
                "I0_net_counts": [0],
                "aerotech_horiz": -800.0,
                "aerotech_horiz_user_setpoint": -800.0,
                "motor": 119.983,
                "motor_setpoint": 120.0,
            },
            "timestamps": {
                "I0_net_counts": [1691957269.0734286],
                "aerotech_horiz": 1691957266.1137164,
                "aerotech_horiz_user_setpoint": 1691957266.1137164,
                "motor": 1692072398.879956,
                "motor_setpoint": 1692072398.8799553,
            },
            "time": 1691957269.0734286,
        },
    ]
    assert len(events) == 2
    assert events == expected_events


@pytest.mark.skip(reason="grid scans are currently broken")
def test_fly_grid_scan(aerotech_flyer):
    flyer = aerotech_flyer
    stepper = sim.motor
    # step size == 10
    plan = grid_fly_scan(
        [], stepper, -100, 100, 11, flyer, -20, 30, 6, snake_axes=[flyer]
    )
    messages = list(plan)
    assert messages[0].command == "stage"
    assert messages[1].command == "open_run"
    # Check that we move the stepper first
    assert messages[2].command == "checkpoint"
    assert messages[3].command == "set"
    assert messages[3].args == (-100,)
    assert messages[4].command == "wait"
    # Check that flyer motor positions snake back and forth
    stepper_positions = [
        msg.args[0]
        for msg in messages
        if (msg.command == "set" and msg.obj.name == "motor")
    ]
    flyer_start_positions = [
        msg.args[0]
        for msg in messages
        if (
            msg.command == "set"
            and msg.obj.name == f"{flyer.name}_flyer_start_position"
        )
    ]
    flyer_end_positions = [
        msg.args[0]
        for msg in messages
        if (msg.command == "set" and msg.obj.name == f"{flyer.name}_flyer_end_position")
    ]
    assert stepper_positions == list(np.linspace(-100, 100, num=11))
    assert flyer_start_positions == [-20, 30, -20, 30, -20, 30, -20, 30, -20, 30, -20]
    assert flyer_end_positions == [30, -20, 30, -20, 30, -20, 30, -20, 30, -20, 30]


@pytest.mark.skip(reason="grid scans are currently broken")
def test_fly_grid_scan_metadata(sim_registry, aerotech_flyer, sim_ion_chamber):
    """Does the plan set the parameters of the flyer motor."""
    flyer = aerotech_flyer
    stepper = sim.motor
    md = {"spam": "eggs"}
    plan = grid_fly_scan(
        [sim_ion_chamber],
        stepper,
        -100,
        100,
        11,
        flyer,
        -20,
        30,
        6,
        snake_axes=[flyer],
        md=md,
    )
    # Check the metadata contained in the "open_run" message
    messages = list(plan)
    open_msg = messages[2]
    assert open_msg.command == "open_run"
    real_md = open_msg.kwargs
    expected_md = {
        "detectors": ["I00"],
        "motors": ("motor", flyer.name),
        "num_points": 66,
        "num_intervals": 65,
        "plan_args": {
            "detectors": [repr(sim_ion_chamber)],
            "args": [repr(stepper), -100, 100, 11, repr(flyer), -20, 30, 6],
        },
        "plan_name": "grid_fly_scan",
        "hints": {
            "gridding": "rectilinear",
            "dimensions": [(["motor"], "primary"), ([flyer.name], "primary")],
        },
        "shape": (11, 6),
        "extents": ([-100, 100], [-20, 30]),
        "snaking": (False, True),
        "plan_pattern": "outer_product",
        "plan_pattern_args": {
            "args": [
                repr(stepper),
                -100,
                100,
                11,
            ]
        },
        "plan_pattern_module": "bluesky.plan_patterns",
        "spam": "eggs",
    }
    assert real_md == expected_md


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""The Haven beamline control system.

See https://haven-spc.readthedocs.io for full documentation."""

__all__ = [
    "energy_scan",
    "xafs_scan",
]

# Get installed version number
import importlib.metadata

try:
    __version__ = importlib.metadata.version("haven-spc")
except importlib.metadata.PackageNotFoundError:
    import pkg_resources

    __version__ = pkg_resources.get_distribution("haven-spc").version
    del pkg_resources
finally:
    del importlib


# Determine the file path of the ipython startup file
from pathlib import Path

ipython_startup_file = Path(__file__).parent / "ipython_startup.ipy"
del Path


# Force ophyd to use caproto as its backend
# import ophyd
# ophyd.set_cl("caproto")

from . import plans  # noqa: F401
from ._iconfig import load_config  # noqa: F401

#  Top-level imports
# from .catalog import load_catalog, load_data, load_result, tiled_client  # noqa: F401
from .catalog import load_catalog, tiled_client  # noqa: F401
from .constants import edge_energy  # noqa: F401
from .devices import IonChamber, Monochromator, Robot, ion_chamber  # noqa: F401
from .devices.motor import HavenMotor  # noqa: F401
from .energy_ranges import ERange, KRange, merge_ranges  # noqa: F401
from .instrument import beamline  # noqa: F401
from .motor_position import (  # noqa: F401
    get_motor_position,
    list_current_motor_positions,
    list_motor_positions,
    recall_motor_position,
    save_motor_position,
)
from .plans.align_motor import align_motor, align_pitch2  # noqa: F401
from .plans.align_slits import align_slits  # noqa: F401
from .plans.auto_gain import GainRecommender, auto_gain  # noqa:F401
from .plans.beam_properties import fit_step  # noqa: F401
from .plans.beam_properties import knife_scan  # noqa: F401
from .plans.energy_scan import energy_scan  # noqa: F401
from .plans.fly import fly_scan, grid_fly_scan  # noqa: F401
from .plans.record_dark_current import record_dark_current  # noqa: F401
from .plans.robot_transfer_sample import robot_transfer_sample  # noqa: F401
from .plans.set_energy import set_energy  # noqa: F401
from .plans.shutters import close_shutters, open_shutters  # noqa: F401
from .plans.xafs_scan import xafs_scan  # noqa: F401
from .preprocessors import (  # noqa: F401
    baseline_decorator,
    baseline_wrapper,
    shutter_suspend_decorator,
    shutter_suspend_wrapper,
)
from .progress_bar import ProgressBar  # noqa: F401
from .run_engine import run_engine  # noqa: F401

# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



#!/usr/bin/env python3
import contextlib
import importlib
import logging
import os
import signal
import time
from pathlib import Path
from subprocess import PIPE, Popen
from textwrap import dedent
from typing import Any, Dict, List, Optional, Tuple

from caproto import ChannelType
from caproto.server import PVGroup, pvproperty, records, run, template_arg_parser
from epics import caget
from tqdm import tqdm

from . import exceptions

log = logging.getLogger(__name__)


locks = {
    "caproto": "unlocked",
}


# Subclass the motor fields here. It's important to use this 'register_record'
# decorator to tell caproto where to find this record:
@records.register_record
class ResponsiveMotorFields(records.MotorFields):
    # The custom fields are identified by this string, which is overridden from
    # the superclass _record_type of 'motor':
    _record_type = "responsive_motor"

    # To override or extend the motor fields, we have to duplicate them here:
    user_readback_value = pvproperty(
        name="RBV", dtype=ChannelType.DOUBLE, doc="User Readback Value", read_only=True
    )

    # Then we are free to extend the fields as normal pvproperties:
    @user_readback_value.scan(period=0.1)
    async def user_readback_value(self, instance, async_lib):
        setpoint = self.parent.value
        pos = setpoint
        # Set the use, dial, and then raw readbacks:
        timestamp = time.time()
        await instance.write(pos, timestamp=timestamp)
        await self.dial_readback_value.write(pos, timestamp=timestamp)
        await self.raw_readback_value.write(int(pos * 100000.0), timestamp=timestamp)


class IOC(PVGroup):
    @classmethod
    def parse_args(Cls) -> Tuple[dict, dict]:
        ioc_options, run_options = ioc_arg_parser(
            default_prefix=Cls.default_prefix, argv=[], desc=dedent(str(Cls.__doc__))
        )
        ioc = Cls(**ioc_options)
        run_options["log_pv_names"] = True
        # run_options["interfaces"] = ["127.0.0.1"]
        return ioc.pvdb, run_options


def wait_for_ioc(pvdb, process=None, timeout=30):
    """Block until all the PVs in the IOC have loaded."""
    # Build a list of PVs and PV fields
    all_fields = []
    fields_found = []
    for pv_name, prop in pvdb.items():
        all_fields.append(pv_name)
    # Wait until all the PVs have responded
    start_time = time.time()
    deadline = start_time + timeout
    all_done = False
    pbar = tqdm(total=len(all_fields), desc="Loading IOC")
    field_times = {}
    while not all_done:
        fields_left = [f for f in all_fields if f not in fields_found]
        all_done = len(fields_left) == 0
        for field in fields_left:
            val = caget(field, timeout=0.5)
            if val is not None:
                fields_found.append(field)
                pbar.update(1)
                field_times[field] = time.time() - start_time
            # Check for exceeding the timeout
            if time.time() > deadline:
                msg = (
                    f"IOC ({list(pvdb)[0]}) did not start within "
                    f"{timeout} seconds. Missing: {fields_left}"
                )
                if process is not None:
                    log.error(f"IOC output: {process.stdout.read()}")
                pbar.close()
                raise exceptions.IOCTimeout(msg)
        time.sleep(0.1)
    log.debug(f"wait_for_ioc() took {time.time() - start_time:.2f} sec.")
    pbar.close()


def run_ioc(pvdb, **kwargs):
    return run(pvdb=pvdb, **kwargs)


# def simulated_ioc(IOCs, prefixes=[], fp=""):
@contextlib.contextmanager
def simulated_ioc(fp):
    # Determine name of the IOC from filename
    fp = Path(fp)
    name = fp.stem
    locks["caproto"] = name
    # Build the pv database
    spec = importlib.util.spec_from_file_location("ioc", str(fp))
    ioc_mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(ioc_mod)
    pvdb, _ = ioc_mod.IOC.parse_args()
    # Make sure the IOC is not already running
    test_pv = list(pvdb.keys())[0]
    response = caget(test_pv, timeout=1.0, connection_timeout=1.0)
    if response is None:
        # IOC is not running, so start it
        process = Popen(
            ["python", str(fp.resolve())], stdout=PIPE, stderr=PIPE, text=True
        )
    else:
        process = None
        log.warning(f"IOC already running: {test_pv} = {response}")
    # Wait for the ioc to load
    wait_for_ioc(pvdb=pvdb, process=process)
    # Drop into the calling code to run the tests
    yield pvdb
    # Stop the process now that the test is done
    if process is not None:
        start_time = time.time()
        os.kill(process.pid, signal.SIGINT)
        stdout, stderr = process.communicate()
        # print(stdout)
        # print(stderr)
        locks["caproto"] = "unlocked"
        log.debug(f"Shutting down took {time.time() - start_time:.2f} sec.")


def ioc_arg_parser(
    *,
    desc: str,
    default_prefix: str,
    argv: Optional[List[str]] = None,
    macros: Optional[Dict[str, str]] = None,
    supported_async_libs: Optional[List[str]] = None,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """A reusable ArgumentParser for basic example IOCs.

    Copied from caproto.server and adjusted to accept *argv* properly.

    Parameters
    ----------
    description : string
        Human-friendly description of what that IOC does
    default_prefix : string
    argv : list, optional
        Defaults to sys.argv
    macros : dict, optional
        Maps macro names to default value (string) or None (indicating that
        this macro parameter is required).
    supported_async_libs : list, optional
        "White list" of supported server implementations. The first one will
        be the default. If None specified, the parser will accept all of the
        (hard-coded) choices.

    Returns
    -------
    ioc_options : dict
        kwargs to be handed into the IOC init.
    run_options : dict
        kwargs to be handed to run

    """
    parser, split_args = template_arg_parser(
        desc=desc,
        default_prefix=default_prefix,
        argv=argv,
        macros=macros,
        supported_async_libs=supported_async_libs,
    )
    return split_args(parser.parse_args(argv))


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.devices import Table


def test_vertical_table():
    """A table that can only move up and down."""
    tbl = Table(vertical_prefix="255idcVME:m3", name="my_table")
    # Check that the correct components were created
    assert hasattr(tbl, "vertical")
    assert tbl.vertical.user_setpoint.source == "ca://255idcVME:m3.VAL"
    assert not hasattr(tbl, "horizontal")
    assert not hasattr(tbl, "upstream")
    assert not hasattr(tbl, "downstream")
    assert not hasattr(tbl, "pitch")
    assert not hasattr(tbl, "vertical_drive_transform")
    assert not hasattr(tbl, "vertical_readback_transform")


def test_horizontal_table():
    """A table that can only move left and right."""
    tbl = Table(horizontal_prefix="255idVME:m3", name="my_table")
    # Check that the correct components were created
    assert not hasattr(tbl, "vertical")
    assert hasattr(tbl, "horizontal")
    assert tbl.horizontal.user_setpoint.source == "ca://255idVME:m3.VAL"
    assert not hasattr(tbl, "upstream")
    assert not hasattr(tbl, "downstream")
    assert not hasattr(tbl, "pitch")
    assert not hasattr(tbl, "vertical_drive_transform")
    assert not hasattr(tbl, "vertical_readback_transform")


def test_pitch_table():
    """A table that can move vertical and adjust the angle."""
    tbl = Table(
        pseudo_motor_prefix="255idcVME:table_us:",
        transform_prefix="255idcVME:table_us_trans:",
        upstream_prefix="255idcVME:m3",
        downstream_prefix="255idcVME:m4",
        name="my_table",
    )
    # Check that the correct components were created (or not)
    assert not hasattr(tbl, "horizontal")
    assert hasattr(tbl, "upstream")
    assert tbl.upstream.user_readback.source == "ca://255idcVME:m3.RBV"
    assert hasattr(tbl, "downstream")
    assert tbl.downstream.user_readback.source == "ca://255idcVME:m4.RBV"
    assert hasattr(tbl, "vertical")
    assert tbl.vertical.user_readback.source == "ca://255idcVME:table_us:height.RBV"
    assert hasattr(tbl, "pitch")
    assert tbl.pitch.user_readback.source == "ca://255idcVME:table_us:pitch.RBV"
    # Check the transforms
    assert hasattr(tbl, "vertical_drive_transform")
    assert (
        tbl.vertical_drive_transform.units.source
        == "ca://255idcVME:table_us_trans:Drive.EGU"
    )
    assert hasattr(tbl, "vertical_readback_transform")
    assert (
        tbl.vertical_readback_transform.units.source
        == "ca://255idcVME:table_us_trans:Readback.EGU"
    )



import logging
import warnings

from bluesky_queueserver_api import BPlan
from ophydregistry import ComponentNotFound
from pydm.widgets.label import PyDMLabel
from pydm.widgets.line_edit import PyDMLineEdit
from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QDialogButtonBox, QFormLayout, QLineEdit, QVBoxLayout
from xraydb.xraydb import XrayDB

from firefly import display
from haven import beamline

log = logging.getLogger(__name__)


class EnergyCalibrationDialog(QtWidgets.QDialog):
    """A dialog box for calibrating the energy."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("Energy calibration")

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        # Widgets for inputting calibration parameters
        self.form_layout = QFormLayout()
        self.layout.addLayout(self.form_layout)
        self.form_layout.addRow(
            "Energy readback:", PyDMLabel(self, init_channel="haven://energy.readback")
        )
        self.form_layout.addRow(
            "Energy setpoint:",
            PyDMLineEdit(self, init_channel="haven://energy.setpoint"),
        )
        self.form_layout.addRow(
            "Calibrated energy:",
            QLineEdit(),
        )
        # Button for accept/close
        buttons = QDialogButtonBox.Apply | QDialogButtonBox.Close
        self.buttonBox = QDialogButtonBox(buttons)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.layout.addWidget(self.buttonBox)


class EnergyDisplay(display.FireflyDisplay):
    stylesheet_danger = (
        "background: rgb(220, 53, 69); color: white; border-color: rgb(220, 53, 69)"
    )
    stylesheet_normal = ""
    energy_positioner = None

    def __init__(self, args=None, macros={}, **kwargs):
        # Load X-ray database for calculating edge energies
        self.xraydb = XrayDB()
        super().__init__(args=args, macros=macros, **kwargs)

    def customize_device(self):
        try:
            self.energy_positioner = beamline.registry.find("energy")
        except ComponentNotFound:
            warnings.warn("Could not find energy positioner.")
            log.warning("Could not find energy positioner.")

    def set_energy(self, *args, **kwargs):
        energy = float(self.ui.target_energy_lineedit.text())
        log.info(f"Setting new energy: {energy}")
        # Build the queue item
        item = BPlan("set_energy", energy=energy)
        # Submit the item to the queueserver
        self.queue_item_submitted.emit(item)

    def update_queue_status(self, status):
        self.set_energy_button.update_queue_style(status)

    def customize_ui(self):
        self.ui.set_energy_button.clicked.connect(self.set_energy)
        # Set up the combo box with X-ray energies
        combo_box = self.ui.edge_combo_box
        ltab = self.xraydb.tables["xray_levels"]
        edges = self.xraydb.query(ltab)
        # min_energy, max_energy = self.energy_positioner.limits
        min_energy, max_energy = 4000, 33000
        edges = edges.filter(
            ltab.c.absorption_edge < max_energy,
            ltab.c.absorption_edge > min_energy,
        )
        items = [
            f"{r.element} {r.iupac_symbol} ({int(r.absorption_edge)} eV)"
            for r in edges.all()
        ]
        combo_box.addItems(["Select edge…", *items])
        combo_box.activated.connect(self.select_edge)
        # Respond to the "calibrate" button
        self.ui.calibrate_button.clicked.connect(self.show_calibrate_dialog)

    def show_calibrate_dialog(self):
        dialog = EnergyCalibrationDialog(self)
        dialog.exec()

    @QtCore.Slot(int)
    def select_edge(self, index):
        if index == 0:
            # The placeholder text was selected
            return
        # Parse the combo box text to get the selected edge
        combo_box = self.ui.edge_combo_box
        text = combo_box.itemText(index)
        elem, edge = text.replace(" ", "_").split("_")[:2]
        # Determine which energy was selected
        edge_info = self.xraydb.xray_edge(element=elem, edge=edge)
        if edge_info is None:
            # Edge is not recognized, so provide feedback
            combo_box.setStyleSheet(self.stylesheet_danger)
        else:
            # Set the text field to the selected edge's energy
            energy, fyield, edge_jump = edge_info
            self.ui.target_energy_lineedit.setText(f"{energy:.3f}")
            combo_box.setStyleSheet(self.stylesheet_normal)

    def ui_filename(self):
        return "energy.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings
from pathlib import Path

from firefly import display
from haven import beamline, load_config
from haven.devices import Table

log = logging.getLogger(__name__)


class CaQtDMBase:
    """caQtDM parameters for the given table.

    Different table geometries require different caQtDM
    configurations.

    """

    ui_file: str = ""
    table: Table

    def __init__(self, table: Table):
        self.table = table

    def motor_attrs(self):
        attr_names = ["horizontal", "vertical"]
        # Look for which motor is present on this table device
        attrs = []
        for attr in attr_names:
            if getattr(self.table, attr, None) is not None:
                attrs.append(attr)
        return attrs

    @property
    def macros(self):
        # See if the Haven config file has the caQtDM macro string
        config = load_config()
        try:
            macro_str = config["table"][self.table.name]["caqtdm_macros"]
        except KeyError:
            log.warning(f"No caQtDM macro string for table: {self.table.name}")
            return {}
        # Parse out the caQtDM string
        macros = {
            k: v for k, v in (piece.split("=", 1) for piece in macro_str.split(","))
        }
        return macros


class TwoLegCaQtDM(CaQtDMBase):
    """caQtDM parameters for a table with two ."""

    ui_file = "/net/s25data/xorApps/ui/table_2leg.ui"


def parse_motor_source(motor):
    source = motor.user_readback.source
    transport, pv = source.split("://", maxsplit=1)
    prefix, suffix = pv.removesuffix(".RBV").rsplit(":", maxsplit=1)
    return transport, prefix, suffix


class SingleMotorCaQtDM(CaQtDMBase):
    """caQtDM parameters for a table with only a single motor."""

    ui_file = "/APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/motorx.ui"


class MultipleMotorCaQtDM(CaQtDMBase):
    """caQtDM parameters for a table with only multiple independent motors."""

    @property
    def ui_file(self):
        ui_dir = Path(
            "/APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/"
        )
        num_motors = len(self.motor_attrs())
        ui_path = ui_dir / f"motor{num_motors}x.ui"
        return str(ui_path)


class TableDisplay(display.FireflyDisplay):
    def customize_device(self):
        self.device = beamline.registry.find(self.macros()["DEVICE"])
        # Determine which flavor of caQtDM parameters we need
        if self.num_legs == 2:
            self.caqtdm = TwoLegCaQtDM(table=self.device)
        elif self.num_motors == 1:
            self.caqtdm = SingleMotorCaQtDM(table=self.device)
        elif self.num_motors > 1:
            self.caqtdm = MultipleMotorCaQtDM(table=self.device)
        else:
            warnings.warn(
                f"Could not determine caQtDM parameters for device: {self.device}."
            )
            self.caqtdm = CaQtDMBase(table=self.device)

    def ui_filename(self):
        return "table.ui"

    @property
    def caqtdm_ui_file(self):
        return self.caqtdm.ui_file

    @property
    def num_legs(self):
        """How motorized legs does this table have?

        If 1, it's a simple motorized table. If greater than 1, it
        will probably have some angular control over the table
        surface.

        """
        leg_names = {"upstream", "downstream"}
        num_legs = len([name for name in leg_names if hasattr(self.device, name)])
        return num_legs

    @property
    def num_motors(self):
        """How many motors does this table have?

        Does not include the pseudo motors produce through the
        sum2Diff EPICS record.

        """
        motor_names = {"horizontal", "vertical", "upstream", "downstream"}
        num_motors = len([name for name in motor_names if hasattr(self.device, name)])
        return num_motors

    def customize_ui(self):
        # Disable motor controls if the given axis is not available
        self.ui.pitch_embedded_display.setEnabled(hasattr(self.device, "pitch"))
        self.ui.vertical_embedded_display.setEnabled(hasattr(self.device, "vertical"))
        self.ui.horizontal_embedded_display.setEnabled(
            hasattr(self.device, "horizontal")
        )

    def launch_caqtdm(self):
        # Build the macros for the caQtDM panels
        caqtdm_macros = dict(
            **self.caqtdm.macros,
        )
        # Launch the caQtDM panel
        super().launch_caqtdm(macros=caqtdm_macros)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from ophyd import Device, EpicsSignal, EpicsSignalRO
from ophyd import FormattedComponent as FCpt

log = logging.getLogger(__name__)


class NHQ203MChannel(Device):
    """A single channel on a controllable power supply."""

    ch_num: int

    # Device components
    potential = FCpt(
        EpicsSignal,
        name="potential",
        suffix="{prefix}:Volt{ch_num}_rbv",
        write_pv="{prefix}:SetVolt{ch_num}",
        tolerance=2,
    )
    current = FCpt(EpicsSignalRO, name="current", suffix="{prefix}:Curr{ch_num}_rbv")
    ramp_rate = FCpt(
        EpicsSignal,
        name="ramp_rate",
        suffix="{prefix}:RampSpeed{ch_num}",
        write_pv="{prefix}:RampSpeed{ch_num}_rbv",
    )
    status = FCpt(EpicsSignalRO, name="status", suffix="{prefix}:ModStatus{ch_num}_rbv")

    def __init__(
        self,
        prefix: str,
        ch_num: int,
        name: str,
        labels={"power_supplies"},
        *args,
        **kwargs,
    ):
        self.ch_num = ch_num
        super().__init__(prefix=prefix, name=name, labels=labels, *args, **kwargs)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio

import pytest
from ophyd_async.core import set_mock_value
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw, epics_signal_x

from haven.positioner import Positioner


class TestPositioner(Positioner):
    done_value = 1

    def __init__(self, name: str = "", put_complete=False):
        self.setpoint = epics_signal_rw(float, ".VAL")
        self.readback = epics_signal_r(float, ".RBV")
        self.actuate = epics_signal_x("StartC.VAL")
        self.done = epics_signal_r(int, "BusyDeviceM.VAL")
        self.stop_signal = epics_signal_rw(int, "StopC.VAL")
        self.precision = epics_signal_rw(int, ".PREC")
        self.velocity = epics_signal_rw(float, ".VELO")
        self.units = epics_signal_rw(str, ".EGU")
        super().__init__(name=name, put_complete=put_complete)


@pytest.fixture()
async def positioner():
    positioner = TestPositioner()
    await positioner.connect(mock=True)
    await positioner.velocity.set(5)
    return positioner


def test_has_signals(positioner):
    assert hasattr(positioner, "setpoint")
    assert hasattr(positioner, "readback")


async def test_set_with_done_actuate(positioner):
    status = positioner.set(5.3)
    set_mock_value(positioner.done, 0)
    await asyncio.sleep(0.01)  # Let event loop run
    set_mock_value(positioner.done, 1)
    await status


async def test_set_with_readback(positioner):
    """Use the readback value to determine when the move is complete."""
    # Remove the done signal to force monitoring readback
    del positioner.done
    # Do the move
    status = positioner.set(5.3)
    set_mock_value(positioner.readback, 5.3)
    await status


async def test_set_with_put_complete():
    positioner = TestPositioner(put_complete=True)
    await positioner.connect(mock=True)
    await positioner.velocity.set(5)
    # Just make sure it doesn't time-out
    await positioner.set(13)



import pytest
from bluesky_queueserver_api import BPlan
from ophyd import Component as Cpt
from ophyd import Signal
from ophyd.sim import SynAxis
from qtpy import QtCore

from firefly.robot import RobotDisplay


class FakeHavenMotor(SynAxis):
    user_offset = Cpt(Signal, value=0, kind="config")


@pytest.fixture
def sim_motor_registry(sim_registry):
    # Create the motors
    sim_registry.register(FakeHavenMotor(name="motor1"))
    sim_registry.register(FakeHavenMotor(name="motor2"))
    yield sim_registry


@pytest.fixture
async def display(qtbot, sim_motor_registry, robot):
    display = RobotDisplay(macros={"DEVICE": robot.name})
    qtbot.addWidget(display)
    await display.update_devices(sim_motor_registry)
    return display


def test_region_number(display):
    # Check that the display has the right number of rows to start with
    assert display.ui.sample_combo_box.count() == 10
    assert hasattr(display, "regions")
    assert len(display.regions) == 0


@pytest.mark.asyncio
async def test_robot_queued(qtbot, sim_motor_registry, display):
    await display.update_devices(sim_motor_registry)
    display.ui.run_button.setEnabled(True)
    display.ui.num_motor_spin_box.setValue(1)
    await display.update_regions(1)

    # set up a test motor
    display.regions[0].motor_box.combo_box.setCurrentText("motor1")
    display.regions[0].start_line_edit.setText("100")

    expected_item = BPlan("robot_transfer_sample", "robotA", 8, "motor1", 100)

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)



"""Holds ion chamber detector descriptions and assignments to EPICS PVs."""

import asyncio
import logging
import warnings
from typing import Dict

from apstools.utils.misc import safe_ophyd_name
from bluesky.protocols import Triggerable
from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    ConfigSignal,
    StandardReadable,
    TriggerInfo,
    soft_signal_rw,
    wait_for_value,
)

from .labjack import LabJackT7
from .scaler import MultiChannelScaler
from .signal import derived_signal_r
from .srs570 import SRS570PreAmplifier

log = logging.getLogger(__name__)


__all__ = ["IonChamber"]


class IonChamber(StandardReadable, Triggerable):
    """A high-level abstraction of an ion chamber.

    Parameters
    ==========
    auto_name
      If true, or None when no name was provided, the name for
      this motor will be set based on the motor's *description*
      field.


    Sub-Devices
    ===========
    mcs
      A multi-channel scaler pointed to be *scaler_prefix*. This
      device will have two channels, the clock channel (#0) and the
      data channel, determined by *scaler_channel*.

    """

    _ophyd_labels_ = {"ion_chambers", "detectors"}
    _trigger_statuses = {}

    def __init__(
        self,
        scaler_prefix: str,
        scaler_channel: int,
        preamp_prefix: str,
        voltmeter_prefix: str,
        voltmeter_channel: int,
        counts_per_volt_second: float,
        name="",
        auto_name: bool = None,
    ):
        self.scaler_prefix = scaler_prefix
        self._scaler_channel = scaler_channel
        self._voltmeter_channel = voltmeter_channel
        self.auto_name = auto_name
        with self.add_children_as_readables():
            self.preamp = SRS570PreAmplifier(preamp_prefix)
            self.voltmeter = LabJackT7(
                prefix=voltmeter_prefix,
                analog_inputs=[voltmeter_channel],
                digital_ios=[],
                analog_outputs=[],
                digital_words=[],
            )
        with self.add_children_as_readables(ConfigSignal):
            self.counts_per_volt_second = soft_signal_rw(
                float, initial_value=counts_per_volt_second
            )
        # Add subordinate devices
        self.mcs = MultiChannelScaler(
            prefix=scaler_prefix, channels=[0, scaler_channel]
        )
        self.add_readables([self.mcs.scaler])
        self.add_readables(
            [
                self.mcs.acquire_mode,
                self.mcs.channel_1_source,
                self.mcs.channel_advance_source,
                self.mcs.count_on_start,
                self.mcs.dwell_time,
                self.mcs.firmware,
                self.mcs.input_mode,
                self.mcs.input_polarity,
                self.mcs.lne_output_delay,
                self.mcs.lne_output_polarity,
                self.mcs.lne_output_stretcher,
                self.mcs.lne_output_width,
                self.mcs.model,
                self.mcs.mux_output,
                self.mcs.num_channels,
                self.mcs.num_channels_max,
                self.mcs.output_mode,
                self.mcs.output_polarity,
                self.mcs.prescale,
                self.mcs.preset_time,
                self.mcs.snl_connected,
            ],
            ConfigSignal,
        )
        # Add calculated signals
        with self.add_children_as_readables():
            self.net_current = derived_signal_r(
                float,
                name="current",
                units="A",
                derived_from={
                    "gain": self.preamp.gain,
                    "count": self.scaler_channel.net_count,
                    "clock_count": self.mcs.scaler.channels[0].raw_count,
                    "clock_frequency": self.mcs.scaler.clock_frequency,
                    "counts_per_volt_second": self.counts_per_volt_second,
                },
                inverse=self._counts_to_amps,
            )
        # Measured current without dark current correction
        self.raw_current = derived_signal_r(
            float,
            name="current",
            units="A",
            derived_from={
                "gain": self.preamp.gain,
                "count": self.scaler_channel.raw_count,
                "clock_count": self.mcs.scaler.channels[0].raw_count,
                "clock_frequency": self.mcs.scaler.clock_frequency,
                "counts_per_volt_second": self.counts_per_volt_second,
            },
            inverse=self._counts_to_amps,
        )
        super().__init__(name=name)

    def _counts_to_amps(
        self,
        values,
        *,
        count,
        gain,
        clock_count,
        clock_frequency,
        counts_per_volt_second,
    ):
        """Pre-amp output current calculated from scaler counts."""
        try:
            # Calculate the output voltage from the pre-amp
            time = values[clock_count] / values[clock_frequency]
            voltage = values[count] / values[counts_per_volt_second] / time
            # Calculate the input current from pre-amp gain
            return voltage / values[gain]
        except ZeroDivisionError:
            return float("nan")

    def __repr__(self):
        return f"<{type(self).__name__}: '{self.name}' ({self.scaler_channel.raw_count.source})>"

    @property
    def scaler_channel(self):
        return self.mcs.scaler.channels[self._scaler_channel]

    @property
    def voltmeter_channel(self):
        return self.voltmeter.analog_inputs[self._voltmeter_channel]

    @property
    def mca(self):
        return self.mcs.mcas[self._scaler_channel]

    async def connect(
        self,
        mock: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ):
        """Connect self and all child Devices.

        Contains a timeout that gets propagated to child.connect methods.

        Parameters
        ----------
        mock:
            If True then use ``MockSignalBackend`` for all Signals
        timeout:
            Time to wait before failing with a TimeoutError.

        """
        await super().connect(
            mock=mock, timeout=timeout, force_reconnect=force_reconnect
        )
        # Update the device's name
        auto_name = bool(self.auto_name) or (self.auto_name is None and self.name == "")
        if bool(auto_name):
            try:
                desc = await self.scaler_channel.description.get_value()
            except Exception as exc:
                warnings.warn(
                    f"Could not read description for {self}. Name not updated. {exc}"
                )
                return
            # Only update the name if the description has been set
            if desc != "":
                self.set_name(safe_ophyd_name(desc))
                # Update the labjack's input's .DESC field to match the scaler channel
                await self.voltmeter_channel.description.set(desc)

    @AsyncStatus.wrap
    async def trigger(self, record_dark_current=False):
        """Instruct the ion chamber's scaler to capture one data point.

        Parameters
        ==========
        record_dark_current
          If true, this measurement will be saved as the dark current
          reading.

        """
        # Make sure we have a fresh voltmeter reading
        await self.voltmeter_channel.trigger()
        # Recording the dark current is done differently
        if record_dark_current:
            await self.record_dark_current()
            return
        # Check if we've seen this signal before
        signal = self.mcs.scaler.count
        last_status = self._trigger_statuses.get(signal.source)
        # Previous trigger is still going, so wait for that instead
        if last_status is not None and not last_status.done:
            await last_status
            return
        # Nothing to wait on yet, so trigger the scaler and stash the result
        st = signal.set(self.mcs.scaler.CountState.COUNT)
        self._trigger_statuses[signal.source] = st
        await st

    async def record_dark_current(self):
        signal = self.mcs.scaler.record_dark_current
        await signal.trigger(wait=False)
        # Now wait for the count state to return to done
        integration_time = await self.mcs.scaler.dark_current_time.get_value()
        timeout = integration_time + DEFAULT_TIMEOUT
        count_signal = self.mcs.scaler.count
        await wait_for_value(
            count_signal, self.mcs.scaler.CountState.DONE, timeout=timeout
        )

    def record_fly_reading(self, reading, **kwargs):
        if self._is_flying:
            self._fly_readings.append(reading)

    @AsyncStatus.wrap
    async def prepare(self, value: TriggerInfo):
        """Prepare the ion chamber for fly scanning."""
        self.start_timestamp = None
        # Set some configuration PVs on the MCS
        await asyncio.gather(
            self.mcs.count_on_start.set(1),
            self.mcs.channel_advance_source.set(self.mcs.ChannelAdvanceSource.INTERNAL),
            self.mcs.num_channels.set(await self.mcs.num_channels_max.get_value()),
            self.mcs.dwell_time.set(value.livetime),
            self.mcs.erase_all.trigger(),
        )
        # Start acquiring data
        self._fly_readings = []
        self._is_flying = False  # Gets set during kickoff

    @AsyncStatus.wrap
    async def kickoff(self):
        """Start recording data for the fly scan."""
        # Watch for new data being collected so we can save timestamps
        self.mcs.current_channel.subscribe(self.record_fly_reading)
        # Start acquiring
        self.mcs.start_all.trigger(wait=False)
        # Wait for acquisition to start
        await wait_for_value(
            self.mcs.acquiring, self.mcs.Acquiring.ACQUIRING, timeout=DEFAULT_TIMEOUT
        )
        self._is_flying = True
        return

    @AsyncStatus.wrap
    async def complete(self):
        """Finish detector fly-scan acquisition."""
        await self.mcs.stop_all.trigger()
        self._is_flying = False

    async def collect_pages(self):
        # Prepare the individual signal data-sets
        timestamps = [
            d[self.mcs.current_channel.name]["timestamp"] for d in self._fly_readings
        ]
        num_points = len(timestamps)
        raw_counts = (await self.mca.spectrum.get_value())[:num_points]
        raw_times = await self.mcs.mcas[0].spectrum.get_value()
        clock_freq = await self.mcs.scaler.clock_frequency.get_value()
        times = raw_times / clock_freq
        # Apply the dark current correction
        offset_rate = await self.scaler_channel.offset_rate.get_value()
        net_counts = raw_counts - offset_rate * times
        # Build the results dictionary to be sent out
        data = {
            self.scaler_channel.raw_count.name: raw_counts,
            self.scaler_channel.net_count.name: net_counts,
            self.mcs.scaler.elapsed_time.name: times,
        }
        results = {
            "time": max(timestamps),
            "data": data,
            "timestamps": {key: timestamps for key in data.keys()},
        }
        yield results

    async def describe_collect(self) -> Dict[str, Dict]:
        signals = [
            self.scaler_channel.raw_count,
            self.scaler_channel.net_count,
            self.mcs.scaler.elapsed_time,
        ]
        descriptions = await asyncio.gather(*(sig.describe() for sig in signals))
        desc = {k: v for desc in descriptions for k, v in desc.items()}
        return {self.name: desc}

    @property
    def default_time_signal(self):
        """The signal to use for setting exposure time when no other signal is
        provided.

        """
        return self.mcs.scaler.preset_time



import json

from pydm.widgets import PyDMEmbeddedDisplay
from qtpy import QtWidgets

from firefly import display
from haven import beamline


class IocsDisplay(display.FireflyDisplay):
    _ioc_displays = []

    def customize_ui(self):
        # Delete existing IOC widgets
        for idx in reversed(range(self.iocs_layout.count())):
            self.iocs_layout.takeAt(idx).widget().deleteLater()
        # Add embedded displays for all the ion chambers
        self._ioc_displays = []
        manager = beamline.registry["beamline_manager"]

        for idx, cpt_name in enumerate(manager.iocs.component_names):
            cpt = getattr(manager.iocs, cpt_name)
            # Add a separator
            if idx > 0:
                line = QtWidgets.QFrame(self.ui)
                line.setObjectName("line")
                # line->setGeometry(QRect(140, 80, 118, 3));
                line.setFrameShape(QtWidgets.QFrame.HLine)
                line.setFrameShadow(QtWidgets.QFrame.Sunken)
                self.iocs_layout.addWidget(line)
            # Create the display object
            disp = PyDMEmbeddedDisplay(parent=self)
            name = cpt.dotted_name.split(".")[-1].lstrip("ioc")
            disp.macros = json.dumps({"IOC": cpt.name, "NAME": name})
            disp.filename = "ioc.ui"
            # Add the Embedded Display to the Results Layout
            self.iocs_layout.addWidget(disp)
            self._ioc_displays.append(disp)

    def ui_filename(self):
        return "iocs.ui"



import asyncio
import inspect
import numbers
from functools import partial
from typing import Callable, Mapping, Optional, Sequence, Type

import numpy as np
from bluesky.protocols import Reading, Subscribable
from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    ReadingValueCallback,
    SignalMetadata,
    SignalR,
    SignalRW,
    SignalX,
    SoftSignalBackend,
    T,
)


class DerivedSignalBackend(SoftSignalBackend):
    """Links a signal to the values of one or more other signals.

    The argument *derived_from* gives the signals that will be used
    for deriving this signal. It should be a mapping of argument names
    to ophyd-async signals, and will be given as keyword arguments to
    the *inverse* and *forward* transforms.

    The default behavior will forward the set value to the real
    signals, and read the real signal's value back (or average of the
    signals if multiple are given).

    To customize this behavior, provide the *forward* or *inverse*
    arguments when creating this backend, or subclass this backend and
    override the ``forward()`` and ``inverse()`` methods.

    *forward()* should be an async function that accepts a positional
    argument with the value sent to this derived signal, along with
    keyword-only arguments corresponding to the signals indicated in
    *derived_from*. It should return a mapping of real signals
    to their new values.

    *inverse()* should accept a positional argument that is a mapping
    of real signals to their read value, along with
    keyword-only arguments corresponding to the signals indicated in
    *derived_from*. It should return a new value to will be sent to
    the derived signal.

    Parameters
    ==========
    derived_from
      From which other signals does this signal derive. Maps
      transformer arguments names to signals.
    forward
      Transforms the derived signal value to the real signal values.
    inverse
      Transforms the real signal values to the derived signals'
      values.

    """

    def __init__(
        self,
        *args,
        derived_from: Mapping,
        forward: Callable = None,
        inverse: Callable = None,
        **kwargs,
    ):
        self._derived_from = derived_from
        if forward is not None:
            self.forward = forward
        if inverse is not None:
            self.inverse = inverse
        self._cached_readings = {}
        super().__init__(*args, **kwargs)

    async def forward(self, value, **kw):
        """The default forward transform for derived signals.

        This method returns the same value for the real signals as was
        set to the derived signal. This behavior can be overridden
        either by subclassing this backend, or by providing a
        *forward* parameter when creating the backend object.

        """
        # Return the same value for the real signal as the derived signal.
        return {key: value for key in kw.values()}

    def inverse(self, values, **kw):
        """The default inverse transform for derived signals.

        This method returns the same value for the derived signal as
        was set for the real signal. If more than one *derived_from*
        signal was provided, this method will return the median. More
        sophisticated behavior can be specified either by subclassing
        this backend, or by providing a *inverse* parameter when
        creating the backend object.

        """
        # Determine a sensible inverse transform value if possible
        is_numeric = all(isinstance(val, numbers.Number) for val in values.values())
        if is_numeric:
            return np.median(tuple(values.values()))
        elif len(values) == 1:
            # Only one value, so return it as-is
            return list(values.values())[0]
        else:
            # No sensible value is possible
            msg = "Cannot determine inverse value for {self} from {values}. "
            msg += "Provide an explicit inverse transform."
            raise ValueError(msg)

    def source(self, name: str = ""):
        src = super().source(name)
        args = ",".join(self._derived_from.keys())
        return f"{src}({args})"

    async def connect(self, timeout=DEFAULT_TIMEOUT) -> None:
        await super().connect(timeout=timeout)
        # Ensure dependent signals are connected
        connectors = (
            sig.connect(timeout=timeout) for sig in self._derived_from.values()
        )
        await asyncio.gather(*connectors)
        # Listen for changes in the derived_from signals
        for sig in self._derived_from.values():
            # Subscribe with a partial in case the signal's name changes
            if isinstance(sig, Subscribable):
                sig.subscribe(partial(self.update_readings, signal=sig))

    def combine_readings(self, readings):
        timestamp = max([rd["timestamp"] for rd in readings.values()])
        severity = max([rd.get("severity", 0) for rd in readings.values()])
        values = {sig: rdg["value"] for sig, rdg in readings.items()}
        new_value = self.inverse(values, **self._derived_from)
        return self.converter.reading(new_value, timestamp, severity)

    def update_readings(self, reading, signal):
        """Callback receives readings from derived_from signals.

        Stashes them for later recall.

        """
        # Stash this reading
        self._cached_readings.update({signal: reading[signal.name]})
        # Update interested parties if we have a full set of readings
        self.send_latest_reading()

    def send_latest_reading(self):
        """Force this backend to send the latest readings to the callback."""
        # Update interested parties if we have a full set of readings
        readings = self._cached_readings
        missing_signals = [
            sig for sig in self._derived_from.values() if sig not in readings.keys()
        ]
        if len(missing_signals) == 0:
            # We have all the readings, so update the cached values
            new_reading = self.combine_readings(readings)
            if self.callback is not None:
                self.callback(new_reading, new_reading["value"])

    def set_callback(self, callback: Optional[ReadingValueCallback[T]]) -> None:
        super().set_callback(callback)
        self.send_latest_reading()

    async def put(self, value: Optional[T], wait=True, timeout=None):
        write_value = (
            self.converter.write_value(value)
            if value is not None
            else self._initial_value
        )
        # Calculate the derived set points
        new_values = await self.forward(write_value, **self._derived_from)
        # Set the new values
        aws = []
        for sig, val in new_values.items():
            if isinstance(sig, SignalX):
                # SignalX objects can't be set, so it must have been triggered
                aws.append(sig.trigger(wait=wait, timeout=timeout))
            else:
                # Check that the independent signal accepts "wait" args
                params = inspect.signature(sig.set).parameters
                kw = {}
                if "wait" in params:
                    kw["wait"] = wait
                aws.append(sig.set(val, timeout=timeout, **kw))
        await asyncio.gather(*aws)

    async def get_reading(self) -> Reading:
        signals = self._derived_from.values()
        readings = await asyncio.gather(*(sig.read() for sig in signals))
        readings = {sig: reading[sig.name] for (sig, reading) in zip(signals, readings)}
        # Return a proper reading for this derived value
        return self.combine_readings(readings)

    async def get_value(self) -> T:
        # Retrieve current values from signals
        values = {sig: (await sig.get_value()) for sig in self._derived_from.values()}
        new_value = self.inverse(values, **self._derived_from)
        return self.converter.value(new_value)


def derived_signal_rw(
    datatype: Optional[Type[T]],
    *,
    initial_value: Optional[T] = None,
    name: str = "",
    derived_from: Sequence,
    forward: Callable = None,
    inverse: Callable = None,
    units: str | None = None,
    precision: int | None = None,
) -> SignalRW[T]:
    """Creates a signal linked to one or more other signals.

    The argument *derived_from* gives the existing signals that will
    be used for deriving this signal. It should be a mapping of
    argument names to ophyd-async signals, and will be given as
    keyword arguments to the *inverse* and *forward* transforms
    describe below.

    The default behavior will forward the set value to the real
    signals, and read the real signal's value back (or average of the
    signals if multiple are given).

    To customize this behavior, provide the *forward* or *inverse*
    arguments when creating this backend, or subclass this backend and
    override the ``forward()`` and ``inverse()`` methods.

    *forward()* should be an async function that accepts a positional
    argument with the value sent to this derived signal, along with
    keyword-only arguments corresponding to the signals indicated in
    *derived_from*. It should return a mapping of real signals
    to their new values.

    *inverse()* should accept a positional argument that is a mapping
    of real signals to their read value, along with
    keyword-only arguments corresponding to the signals indicated in
    *derived_from*. It should return a new value to will be sent to
    the derived signal.

    Example:

    .. code-block:: python

        async def squareroot(value, *, voltage):
            return {voltage: value**0.5}

        def square(values, *, voltage):
            return values[voltage]**2

        class MyDevice(Device):
            def __init__(self, prefix, name="", **kwargs):
                self.voltage = soft_signal_rw(int)
                self.voltage_squared = derived_signal_rw(
                    int,
                    derive_from={"voltage": self.voltage},
                    foward=squareroot,
                    inverse=square
                )
                super().__init__(name=name, **kwargs)

    Parameters
    ==========
    derived_from
      From which other signals does this signal derive. Maps
      transformer arguments names to signals.
    forward
      Transforms the derived signal value to the real signal values.
    inverse
      Transforms the real signal values to the derived signals'
      values.

    """
    metadata = SignalMetadata(units=units, precision=precision)
    backend = DerivedSignalBackend(
        datatype,
        derived_from=derived_from,
        forward=forward,
        inverse=inverse,
        initial_value=initial_value,
        metadata=metadata,
    )
    signal = SignalRW(backend, name=name)
    return signal


def derived_signal_r(
    datatype: Optional[Type[T]],
    *,
    initial_value: Optional[T] = None,
    name: str = "",
    derived_from: Sequence,
    inverse: Callable = None,
    units: str | None = None,
    precision: int | None = None,
) -> SignalRW[T]:
    """Creates a signal linked to one or more other signals.

    The argument *derived_from* gives the existing signals that will
    be used for deriving this signal. It should be a mapping of
    argument names to ophyd-async signals, and will be given as
    keyword arguments to the *inverse* transform describe below.

    The default behavior will read the real signal's value back (or
    average of the signals if multiple are given).

    To customize this behavior, provide the *inverse*
    arguments when creating this backend, or subclass this backend and
    override the ``inverse()`` methods.

    *inverse()* should accept a positional argument that is a mapping
    of real signals to their read value, along with keyword-only
    arguments corresponding to the signals indicated in
    *derived_from*. It should return a new value to will be sent to
    the derived signal.

    Example:

    .. code-block:: python

        def square(values, *, voltage):
            return values[voltage]**2

        class MyDevice(Device):
            def __init__(self, prefix, name="", **kwargs):
                self.voltage = soft_signal_rw(int)
                self.voltage_squared = derived_signal_r(
                    int,
                    derive_from={"voltage": self.voltage},
                    inverse=square
                )
                super().__init__(name=name, **kwargs)

    Parameters
    ==========
    derived_from
      From which other signals does this signal derive. Maps
      transformer arguments names to signals.
    inverse
      Transforms the real signal values to the derived signals'
      values.

    """
    metadata = SignalMetadata(units=units, precision=precision)
    backend = DerivedSignalBackend(
        datatype,
        derived_from=derived_from,
        inverse=inverse,
        initial_value=initial_value,
        metadata=metadata,
    )
    signal = SignalR(backend, name=name)
    return signal


def derived_signal_x(
    *,
    name: str = "",
    derived_from: Sequence,
    forward: Callable = None,
) -> SignalX:
    """Creates a signal linked to one or more other signals.

    The argument *derived_from* gives the existing signals that will
    be used for deriving this signal. It should be a mapping of
    argument names to ophyd-async signals, and will be given as
    keyword arguments to the *inverse* transform describe below.

    The default behavior will read the real signal's value back (or
    average of the signals if multiple are given).

    To customize this behavior, provide the *inverse*
    arguments when creating this backend, or subclass this backend and
    override the ``inverse()`` methods.

    *inverse()* should accept a positional argument that is a mapping
    of real signals to their read value, along with keyword-only
    arguments corresponding to the signals indicated in
    *derived_from*. It should return a new value to will be sent to
    the derived signal.

    Example:

    .. code-block:: python

        def square(values, *, voltage):
            return values[voltage]**2

        class MyDevice(Device):
            def __init__(self, prefix, name="", **kwargs):
                self.voltage = soft_signal_rw(int)
                self.voltage_squared = derived_signal_r(
                    int,
                    derive_from={"voltage": self.voltage},
                    inverse=square
                )
                super().__init__(name=name, **kwargs)

    Parameters
    ==========
    derived_from
      From which other signals does this signal derive. Maps
      transformer arguments names to signals.
    inverse
      Transforms the real signal values to the derived signals'
      values.

    """
    metadata = SignalMetadata()
    backend = DerivedSignalBackend(
        int,
        derived_from=derived_from,
        forward=forward,
        metadata=metadata,
    )
    signal = SignalX(backend, name=name)
    return signal



import pytest
from ophyd import sim
from ophyd.utils.errors import ReadOnlyError

from haven.devices.shutter import PssShutter, ShutterState


@pytest.fixture()
def shutter(sim_registry):
    shutter = sim.instantiate_fake_device(PssShutter, name="shutter")
    return shutter


def test_shutter_setpoint(shutter):
    """When we open and close the shutter, do the right EPICS signals get
    set?

    """
    shutter.open_signal.sim_put(0)
    shutter.close_signal.sim_put(0)
    # Close the shutter
    shutter.open_signal.sim_put(0)
    shutter.close_signal.sim_put(0)
    status = shutter.set(ShutterState.CLOSED)
    shutter.readback.sim_put(ShutterState.CLOSED)
    status.wait(timeout=1)
    assert shutter.open_signal.get() == 0
    assert shutter.close_signal.get() == 1
    # Open the shutter
    shutter.close_signal.sim_put(0)
    shutter.open_signal.sim_put(0)
    status = shutter.set(ShutterState.OPEN)
    shutter.readback.sim_put(ShutterState.OPEN)
    status.wait(timeout=1)
    assert shutter.open_signal.get() == 1
    assert shutter.close_signal.get() == 0


def test_shutter_check_value(shutter):
    # Check for non-sense values
    with pytest.raises(ValueError):
        shutter.set(ShutterState.FAULT)
    # Test shutter allow_close
    shutter.allow_close = False
    with pytest.raises(ReadOnlyError):
        shutter.set(ShutterState.CLOSED)
    # Test shutter allow_open
    shutter.allow_open = False
    with pytest.raises(ReadOnlyError):
        shutter.set(ShutterState.OPEN)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DCpt
from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO

log = logging.getLogger(__name__)


LOAD_TIMEOUT = 80


class Sample(Device):
    """An individual robot sample that can be loaded.

    Signals
    =======
    present
      Whether or not a sample is physically present on the stage.
    empty
      Whether or not no sample is physically present on the stage.
    load
      Direct the robot to physically move this sample to the loading
      position. Can be slow (~25-30 seconds).
    unload
      Direct the robot to physically remove this sample to the loading
      position. Can be slow (~25-30 seconds).
    x
      The x position of the robot in Cartesian coordinates.
    y
      The y position of the robot in Cartesian coordinates.
    z
      The z position of the robot in Cartesian coordinates.
    rx
      The rx position of the robot in Cartesian coordinates.
    ry
      The ry position of the robot in Cartesian coordinates.
    rz
      The position of the robot in Cartesian coordinates.

    """

    present = Cpt(EpicsSignalRO, ":present")
    empty = Cpt(EpicsSignalRO, ":empty")
    load = Cpt(
        EpicsSignal,
        ":load",
        kind="omitted",
        write_timeout=LOAD_TIMEOUT,
        put_complete=True,
    )
    unload = Cpt(
        EpicsSignal,
        ":unload",
        kind="omitted",
        write_timeout=LOAD_TIMEOUT,
        put_complete=True,
    )
    x = Cpt(EpicsSignalRO, ":x")
    y = Cpt(EpicsSignalRO, ":y")
    z = Cpt(EpicsSignalRO, ":z")
    rx = Cpt(EpicsSignalRO, ":rx")
    ry = Cpt(EpicsSignalRO, ":ry")
    rz = Cpt(EpicsSignalRO, ":rz")

    def __init__(self, *args, labels={"robots"}, **kwargs):
        super().__init__(*args, labels=labels, **kwargs)


def transfer_samples(num_samples: int):
    """Create a dictionary with robot sample device definitions.
    For use with an ophyd DynamicDeviceComponent.
    Parameters
    ==========
    num_dios
      How many samples to create.
    """
    samples = {}
    # Now the sample holder bases are only located at sites [8,9,10,14,15,16,20,21,22] on the board.
    for n in [8, 9, 10, 14, 15, 16, 20, 21, 22]:  # range(num_samples):
        samples[f"sample{n}"] = (Sample, f":sample{n}", {})
    return samples


class Robot(Device):
    # joints and position
    i = Cpt(EpicsMotor, ":i")
    j = Cpt(EpicsMotor, ":j")
    k = Cpt(EpicsMotor, ":k")
    l = Cpt(EpicsMotor, ":l")
    m = Cpt(EpicsMotor, ":m")
    n = Cpt(EpicsMotor, ":n")
    x = Cpt(EpicsMotor, ":x")
    y = Cpt(EpicsMotor, ":y")
    z = Cpt(EpicsMotor, ":z")
    rx = Cpt(EpicsMotor, ":rx")
    ry = Cpt(EpicsMotor, ":ry")
    rz = Cpt(EpicsMotor, ":rz")
    acc = Cpt(EpicsSignal, ":acceleration", kind="config")
    vel = Cpt(EpicsSignal, ":velocity", kind="config")

    # dashboard
    remote_control = Cpt(EpicsSignalRO, ":dashboard:remote_control", kind="config")
    program = Cpt(EpicsSignal, ":dashboard:program_rbv", kind="omitted")
    program_rbv = Cpt(EpicsSignalRO, ":dashboard:program_rbv", kind="config")
    installation = Cpt(EpicsSignal, ":dashboard:installation", kind="config")
    playRbt = Cpt(EpicsSignal, ":dashboard:play", kind="omitted")
    stopRbt = Cpt(EpicsSignal, ":dashboard:stop", kind="omitted")
    pauseRbt = Cpt(EpicsSignal, ":dashboard:pause", kind="omitted")
    quitRbt = Cpt(EpicsSignal, ":dashboard:quit", kind="omitted")
    shutdown = Cpt(EpicsSignal, ":dashboard:shutdown", kind="omitted")
    release_brake = Cpt(EpicsSignal, ":dashboard:release_brake", kind="omitted")
    close_safety_popup = Cpt(
        EpicsSignal, ":dashboard:close_safety_popup", kind="config"
    )
    unlock_protective_stop = Cpt(
        EpicsSignal, ":dashboard:unlock_protective_stop", kind="config"
    )
    restart_safety = Cpt(EpicsSignal, ":dashboard:restart_safety", kind="config")
    program_running = Cpt(EpicsSignal, ":dashboard:program_running", kind="config")
    safety_status = Cpt(EpicsSignal, ":dashboard:safety_status", kind="config")
    power = Cpt(EpicsSignal, ":dashboard:power", kind="omitted")
    power_rbv = Cpt(EpicsSignalRO, ":dashboard:power_rbv", kind="config")

    # gripper
    gripper_activate = Cpt(EpicsSignal, ":gripper.ACT", kind="omitted")
    gripper_activated = Cpt(EpicsSignal, ":gripper.ACR", kind="config")
    gripper_close = Cpt(EpicsSignal, ":gripper.CLS", kind="omitted")
    gripper_open = Cpt(EpicsSignal, ":gripper.OPN", kind="omitted")
    gripper_rbv = Cpt(EpicsSignal, ":gripper.RBV")
    gripper_val = Cpt(EpicsSignal, ":gripper.VAL")
    gripper_force = Cpt(EpicsSignal, ":gripper.FRC", kind="config")

    # busy
    busy = Cpt(EpicsSignal, ":busy", kind="omitted")

    # sample transfer
    current_sample = Cpt(EpicsSignalRO, ":current_sample", kind="config")
    unload_current_sample = Cpt(
        EpicsSignal,
        ":unload_current_sample",
        kind="omitted",
        write_timeout=LOAD_TIMEOUT,
        put_complete=True,
    )
    current_sample_reset = Cpt(EpicsSignal, ":current_sample_reset", kind="omitted")
    home = Cpt(EpicsSignal, ":home", kind="config")
    cal_stage = Cpt(EpicsSignal, ":cal_stage", kind="config")

    samples = DCpt(transfer_samples(24))


# -----------------------------------------------------------------------------
# :author:    Yanna Chen
# :email:     yannachen@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Test the Labjack T-series data acquisition device support

Hardware is not available so test with best efforts

"""

import pytest

from haven.devices.labjack import (
    AnalogInput,
    AnalogOutput,
    DigitalIO,
    LabJackBase,
    LabJackT4,
    LabJackT7,
    LabJackT7Pro,
    LabJackT8,
    WaveformDigitizer,
    WaveformGenerator,
)

PV_PREFIX = "255idc:LJ_T7:"


@pytest.fixture()
async def labjack():
    lj = LabJackBase(PV_PREFIX, name="labjack", analog_outputs=range(2))
    await lj.connect(mock=True)
    return lj


@pytest.mark.asyncio
async def test_base_signals_device(labjack):
    """Test signals shared by all labjack devices."""
    cfg_names = {
        "labjack-model_name",
        "labjack-firmware_version",
        "labjack-serial_number",
        "labjack-device_temperature",
        "labjack-ljm_version",
        "labjack-driver_version",
        "labjack-last_error_message",
        "labjack-poll_sleep_ms",
        "labjack-analog_in_settling_time_all",
        "labjack-analog_in_resolution_all",
        "labjack-analog_in_sampling_rate",
        "labjack-analog_outputs-0-description",
        "labjack-analog_outputs-0-desired_output_location",
        "labjack-analog_outputs-0-device_type",
        "labjack-analog_outputs-0-output_link",
        "labjack-analog_outputs-0-output_mode_select",
        "labjack-analog_outputs-0-scanning_rate",
        "labjack-analog_outputs-1-description",
        "labjack-analog_outputs-1-desired_output_location",
        "labjack-analog_outputs-1-device_type",
        "labjack-analog_outputs-1-output_link",
        "labjack-analog_outputs-1-output_mode_select",
        "labjack-analog_outputs-1-scanning_rate",
    }
    desc = await labjack.describe_configuration()
    from pprint import pprint

    pprint(await labjack.analog_outputs[0].describe_configuration())
    # pprint(desc)
    assert set(desc.keys()) == cfg_names


ai_params = [
    # (model, number of analog inputs)
    (LabJackT4, 12),
    (LabJackT7, 14),
    (LabJackT7Pro, 14),
    (LabJackT8, 8),
]


@pytest.mark.parametrize("LabJackDevice,num_ais", ai_params)
async def test_analog_inputs(LabJackDevice, num_ais):
    """Test analog inputs for different device types."""
    device = LabJackDevice(PV_PREFIX, name="labjack_T")
    await device.connect(mock=True)
    assert hasattr(device, "analog_inputs")
    # Check that the individual AI devices were created
    for n in range(num_ais):
        assert n in device.analog_inputs.keys()
        ai = device.analog_inputs[n]
        assert isinstance(ai, AnalogInput)
    # Make sure there aren't any extra analog inputs
    assert num_ais not in device.analog_inputs.keys()
    # Check read attrs
    read_attrs = ["final_value"]
    description = await device.describe()
    print(description)
    for n in range(num_ais):
        for attr in read_attrs:
            full_attr = f"{device.name}-analog_inputs-{n}-{attr}"
            assert full_attr in description.keys()
    # Check configuration attrs
    cfg_attrs = [
        "differential",
        "high",
        "low",
        "temperature_units",
        "resolution",
        "range",
        "mode",
        "enable",
    ]
    description = await device.describe_configuration()
    for n in range(num_ais):
        for attr in cfg_attrs:
            full_attr = f"{device.name}-analog_inputs-{n}-{attr}"
            assert full_attr in description.keys()


ao_params = [
    # (model, number of analog outputs)
    (LabJackT4, 2),
    (LabJackT7, 2),
    (LabJackT7Pro, 2),
    (LabJackT8, 2),
]


@pytest.mark.parametrize("LabJackDevice,num_aos", ao_params)
async def test_analog_outputs(LabJackDevice, num_aos):
    """Test analog inputs for different device types."""
    device = LabJackDevice(PV_PREFIX, name="labjack_T")
    await device.connect(mock=True)
    assert hasattr(device, "analog_outputs")
    # Check that the individual AI devices were created
    for n in range(num_aos):
        assert n in device.analog_outputs.keys()
        ai = device.analog_outputs[n]
        assert isinstance(ai, AnalogOutput)
    # Check read attrs
    read_attrs = ["desired_value"]
    description = await device.describe()
    for n in range(num_aos):
        for attr in read_attrs:
            full_attr = f"{device.name}-analog_outputs-{n}-{attr}"
            assert full_attr in description.keys()
    # Check configuration attrs
    cfg_attrs = ["output_mode_select"]
    description = await device.describe_configuration()
    for n in range(num_aos):
        for attr in cfg_attrs:
            full_attr = f"{device.name}-analog_outputs-{n}-{attr}"
            assert full_attr in description.keys()
    # Check hinted attrs
    hinted_attrs = ["readback_value"]
    for n in range(num_aos):
        for attr in hinted_attrs:
            full_attr = f"{device.name}-analog_outputs-{n}-{attr}"
            assert full_attr in device.hints["fields"]


dio_params = [
    # (model, number of digital I/Os)
    (LabJackT4, 16),
    (LabJackT7, 23),
    (LabJackT7Pro, 23),
    (LabJackT8, 20),
]


@pytest.mark.parametrize("LabJackDevice,num_dios", dio_params)
async def test_digital_ios(LabJackDevice, num_dios):
    """Test analog inputs for different device types."""
    device = LabJackDevice(PV_PREFIX, name="labjack_T")
    await device.connect(mock=True)
    assert hasattr(device, "digital_ios")
    # Check that the individual digital I/O devices were created
    for n in range(num_dios):
        assert n in device.digital_ios.keys()
        dio = device.digital_ios[n]
        assert isinstance(dio, DigitalIO)
    # Check read attrs
    read_attrs = ["output-desired_value", "input-final_value"]
    description = await device.describe()
    for n in range(num_dios):
        for attr in read_attrs:
            full_attr = f"{device.name}-digital_ios-{n}-{attr}"
            assert full_attr in description.keys()
    # Check configuration attrs
    cfg_attrs = ["direction"]
    description = await device.describe_configuration()
    for n in range(num_dios):
        for attr in cfg_attrs:
            full_attr = f"{device.name}-digital_ios-{n}-{attr}"
            assert full_attr in description.keys()
    # # Check hinted attrs
    # hinted_attrs = ["output-readback_value"]
    # for n in range(num_dios):
    #     for attr in hinted_attrs:
    #         full_attr = f"{device.name}-digital_ios-{n}-{attr}"
    #         assert full_attr in device.hints["fields"]


@pytest.mark.parametrize("LabJackDevice,num_dios", dio_params)
async def test_digital_words(LabJackDevice, num_dios):
    """Test analog inputs for different device types."""
    device = LabJackDevice(PV_PREFIX, name="labjack_T")
    await device.connect(mock=True)
    assert hasattr(device, "digital_words")
    # Check that the individual digital word outputs were created
    assert "dio" in device.digital_words.keys()
    assert "fio" in device.digital_words.keys()
    assert "eio" in device.digital_words.keys()
    assert "cio" in device.digital_words.keys()
    assert "mio" in device.digital_words.keys()
    # Check read attrs
    read_attrs = ["dio", "eio", "fio", "mio", "cio"]
    description = await device.describe()
    for attr in read_attrs:
        assert f"{device.name}-digital_words-{attr}" in description.keys()


async def test_waveform_digitizer():
    digitizer = WaveformDigitizer("LabJackT7_1:", name="labjack")
    await digitizer.connect(mock=True)
    # Check read attrs
    read_attrs = ["timebase_waveform", "dwell_actual", "total_time"]
    description = await digitizer.describe()
    for attr in read_attrs:
        assert f"{digitizer.name}-{attr}" in description.keys()
    # Check read attrs
    cfg_attrs = [
        "num_points",
        "first_chan",
        "num_chans",
        "dwell_time",
        "resolution",
        "settling_time",
    ]
    description = await digitizer.describe_configuration()
    for attr in cfg_attrs:
        assert f"{digitizer.name}-{attr}" in description.keys()


@pytest.mark.parametrize("LabJackDevice,num_ais", ai_params)
async def test_waveform_digitizer_waveforms(LabJackDevice, num_ais):
    """Verify that the waveform digitizer is created for each LabJack."""
    device = LabJackDevice(PV_PREFIX, name="labjack_T")
    await device.connect(mock=True)
    assert hasattr(device, "waveform_digitizer")
    digitizer = device.waveform_digitizer
    assert hasattr(digitizer, "waveforms")


async def test_waveform_generator():
    generator = WaveformGenerator("LabJackT7_1:", name="labjack")
    await generator.connect(mock=True)
    # Check read attrs
    read_attrs = [
        "frequency",
        "dwell",
        "dwell_actual",
        "total_time",
        "user_time_waveform",
        "internal_time_waveform",
    ]
    description = await generator.describe()
    for attr in read_attrs:
        assert f"{generator.name}-{attr}" in description.keys()
    # Check read attrs
    cfg_attrs = [
        "external_trigger",
        "external_clock",
        "continuous",
        "num_points",
        "user_waveform_0",
        "enable_0",
        "type_0",
        "pulse_width_0",
        "amplitude_0",
        "offset_0",
        "user_waveform_1",
        "enable_1",
        "type_1",
        "pulse_width_1",
        "amplitude_1",
        "offset_1",
    ]
    description = await generator.describe_configuration()
    for attr in cfg_attrs:
        assert f"{generator.name}-{attr}" in description.keys()



"""Bluesky plans specific to spectroscopy."""

# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
import logging
from functools import partial
import warnings

import numpy as np
from bluesky.protocols import Movable, Stoppable
from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    StandardReadable,
    WatchableAsyncStatus,
    WatcherUpdate,
    observe_value,
)

log = logging.getLogger(__name__)


class Positioner(StandardReadable, Movable, Stoppable):
    """A positioner that has separate setpoint and readback signals.

    When set, the Positioner **monitors the state of the move** using
    a strategy selected based on the Positioner's configuration
    (chosen in order):

    1. If *put_complete* is true, await the setpoint's ``set()``
       status to report being done.
    2. If the positioner has a *done* signal attribute, wait for this
       signal to reach the value of ``Positioner.done_value``.
    3. Wait for the readback to be close to the setpoint (following
       :py:func:`numpy.isclose`).

    Parameters
    ==========
    name
      The device name for this positioner.
    put_complete
      If true, wait on the setpoint to report being done.

    """

    done_value = 1

    def __init__(self, name: str = "", put_complete: bool = False):
        self.put_complete = put_complete
        super().__init__(name=name)

    def set_name(self, name: str):
        super().set_name(name)
        # Readback should be named the same as its parent in read()
        self.readback.set_name(name)

    def watch_done(
        self, value, done_event: asyncio.Event, started_event: asyncio.Event
    ):
        """Update the event when the done value is actually done."""
        log.debug(f"Received new done value: {value}.")
        if value != self.done_value:
            # The movement has started
            log.debug("Setting started_event")
            started_event.set()
        elif started_event.is_set():
            # Move has finished
            log.debug("Setting done_event")
            done_event.set()

    @WatchableAsyncStatus.wrap
    async def set(self, value: float, timeout: CalculatableTimeout = CALCULATE_TIMEOUT):
        new_position = value
        self._set_success = True
        old_position, units, precision, velocity = await asyncio.gather(
            self.setpoint.get_value(),
            self.units.get_value(),
            self.precision.get_value(),
            self.velocity.get_value(),
        )
        if timeout == CALCULATE_TIMEOUT:
            assert velocity > 0, "Mover has zero velocity"
            timeout = abs(new_position - old_position) / velocity + DEFAULT_TIMEOUT
        # Make an Event that will be set on completion, and a Status that will
        # error if not done in time
        reached_setpoint = asyncio.Event()
        done_event = asyncio.Event()
        started_event = asyncio.Event()
        # Start the move
        if hasattr(self, "actuate"):
            # Set the setpoint, then click "go"
            await self.setpoint.set(new_position, wait=True)
            set_status = self.actuate.trigger(wait=self.put_complete, timeout=timeout)
        else:
            # Wait for the value to set, but don't wait for put completion callback
            set_status = self.setpoint.set(
                new_position, wait=self.put_complete, timeout=timeout
            )
        # Decide on how we will wait for completion
        if self.put_complete:
            # await the set call directly
            done_status = set_status
        elif hasattr(self, "done"):
            # Monitor the `done` signal
            log.debug(f"Monitoring progress via ``done`` signal: {self.done.name}.")
            self.done.subscribe_value(
                partial(
                    self.watch_done, done_event=done_event, started_event=started_event
                )
            )
            done_status = AsyncStatus(asyncio.wait_for(done_event.wait(), timeout))
        else:
            # Monitor based on readback position
            done_status = AsyncStatus(
                asyncio.wait_for(reached_setpoint.wait(), timeout)
            )
        # Monitor the position of the readback value
        async for current_position in observe_value(
            self.readback, done_status=done_status
        ):
            yield WatcherUpdate(
                current=current_position,
                initial=old_position,
                target=new_position,
                name=self.name,
                unit=units,
                precision=int(precision),
            )
            # Check if the move has finished
            target_reached = current_position is not None and np.isclose(
                current_position, new_position
            )
            if target_reached:
                reached_setpoint.set()
                break
        # Make sure the done point was actually reached
        await done_status
        # Handle failed moves
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    async def stop(self, success=True):
        self._set_success = success
        if hasattr(self, "stop_signal"):
            await self.stop_signal.trigger()
        else:
            warnings.warn(f"Positioner {self.name} has no stop signal.")



from pprint import pprint
from unittest import mock

import numpy as np
import pytest
from bluesky_queueserver_api import BPlan
from qtpy import QtCore

from firefly.plans.xafs_scan import XafsScanDisplay

# default values for EXAFS scan
pre_edge = [-50, -20, 1]
XANES_region = [-20, 50, 0.2]
EXAFS_region = [50, 500, 0.5]
default_values = [pre_edge, XANES_region, EXAFS_region]


@pytest.fixture()
def display(qtbot):
    display = XafsScanDisplay()
    qtbot.addWidget(display)
    return display


def test_region_number(display):
    """Does changing the region number affect the UI?"""
    # Check that the display has the right number of rows to start with
    assert display.ui.regions_spin_box.value() == 3
    assert hasattr(display, "regions")
    assert len(display.regions) == 3

    # Check that regions can be inserted
    display.ui.regions_spin_box.setValue(5)
    assert len(display.regions) == 5

    # Check that regions can be removed
    display.ui.regions_spin_box.setValue(1)
    assert len(display.regions) == 1


def test_E0_checkbox(display):
    """Does selecting the E0 checkbox adjust the UI properly?"""
    # check whether extracted edge value is correct
    display.edge_combo_box.setCurrentText("Pt L3 (11500.8 eV)")
    display.ui.use_edge_checkbox.setChecked(True)

    # check whether the math is done correctly when switching off E0
    display.ui.use_edge_checkbox.setChecked(False)
    # check whether edge value is extracted correctly
    np.testing.assert_equal(display.edge_value, 11500.8)
    # K-space checkboxes should be disabled when E0 is unchecked
    assert not display.regions[0].k_space_checkbox.isEnabled()

    # check whether energy values is added correctly
    for i in range(len(default_values)):
        np.testing.assert_almost_equal(
            float(display.regions[i].start_line_edit.text()),
            default_values[i][0] + display.edge_value,
            decimal=3,
        )
        np.testing.assert_almost_equal(
            float(display.regions[i].stop_line_edit.text()),
            default_values[i][1] + display.edge_value,
            decimal=3,
        )
        np.testing.assert_almost_equal(
            float(display.regions[i].step_line_edit.text()),
            default_values[i][2],
            decimal=3,
        )

    # check whether k range is calculated right
    display.ui.use_edge_checkbox.setChecked(True)
    # K-space checkbox should become re-enabled after E0 is checked
    assert display.regions[-1].k_space_checkbox.isEnabled()
    display.regions[-1].k_space_checkbox.setChecked(True)
    np.testing.assert_almost_equal(
        float(display.regions[i].start_line_edit.text()), 3.6226, decimal=4
    )
    np.testing.assert_almost_equal(
        float(display.regions[i].stop_line_edit.text()), 11.4557, decimal=4
    )
    np.testing.assert_almost_equal(
        float(display.regions[i].step_line_edit.text()), 3.64069 - 3.6226, decimal=4
    )


def test_xafs_scan_plan_queued_energies(display, qtbot):
    display.edge_combo_box.setCurrentText("Pt L3 (11500.8 eV)")
    display.regions[-1].region_checkbox.setChecked(False)
    # Set up detector list
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", "I0"]
    )
    energies_region0 = np.arange(
        default_values[0][0],
        default_values[0][1] + default_values[0][2],
        default_values[0][2],
    )
    energies_region1 = np.arange(
        default_values[1][0] + default_values[1][2],
        default_values[1][1] + default_values[1][2],
        default_values[1][2],
    )
    energies_merge = np.hstack([energies_region0, energies_region1])
    exposures = np.ones(energies_merge.shape)

    # set up meta data
    display.ui.lineEdit_sample.setText("sam")
    display.ui.lineEdit_purpose.setText("test")
    display.ui.checkBox_is_standard.setChecked(True)
    display.ui.lineEdit_purpose.setText("test")
    display.ui.textEdit_notes.setText("sam_notes")

    expected_item = BPlan(
        "energy_scan",
        energies=energies_merge,
        exposure=exposures,
        E0=11500.8,
        detectors=["vortex_me4", "I0"],
        md={
            "sample": "sam",
            "purpose": "test",
            "is_standard": True,
            "notes": "sam_notes",
        },
    )

    def check_item(item):
        item_dict = item.to_dict()["kwargs"]
        expected_dict = expected_item.to_dict()["kwargs"]

        try:
            # Check energies & exposures within 3 decimals
            np.testing.assert_array_almost_equal(
                item_dict["energies"], expected_dict["energies"], decimal=3
            )
            np.testing.assert_array_almost_equal(
                item_dict["exposure"], expected_dict["exposure"], decimal=3
            )

            # Now check the rest of the dictionary, excluding the numpy array keys
            item_dict.pop("energies")
            item_dict.pop("exposure")
            expected_dict.pop("energies")
            expected_dict.pop("exposure")

            # Check if the remaining dictionary items are equal
            assert item_dict == expected_dict, "Non-array items do not match."

        except AssertionError as e:
            # Print detailed debug info
            pprint(item_dict)
            pprint(expected_dict)
            print(str(e))
            return False
        return True

    # Click the run button and see if the plan is queued
    display.ui.run_button.setEnabled(True)
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)


def test_xafs_scan_plan_queued_energies_k_mixed(qtbot, display):
    display.ui.regions_spin_box.setValue(2)
    display.edge_combo_box.setCurrentText("Pt L3 (11500.8 eV)")

    # set up the first region
    display.regions[0].start_line_edit.setText("-20")
    display.regions[0].stop_line_edit.setText("40")
    display.regions[0].step_line_edit.setText("10")

    # set up the second region
    display.regions[1].start_line_edit.setText("50")
    display.regions[1].stop_line_edit.setText("800")

    # convert to k space
    display.regions[1].k_space_checkbox.setChecked(True)
    display.regions[1].step_line_edit.setText("5")
    display.regions[1].weight_spinbox.setValue(2)

    # set up detector list
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", "I0"]
    )

    # set repeat scan num to 2
    display.ui.spinBox_repeat_scan_num.setValue(3)

    energies = np.array(
        [
            -20,
            -10,
            0,
            10,
            20,
            30,
            40,
            50,
            283.27,
            707.04,
        ]  # k values obtained from Athena software to double confirm ours
    )
    exposures = np.array(
        [1, 1, 1, 1, 1, 1, 1, 1, 5.665, 14.141]
    )  # k exposures kmin 3.62263

    # set up meta data
    display.ui.lineEdit_sample.setText("sam")
    display.ui.lineEdit_purpose.setText("")  # invalid input should be removed from md
    display.ui.textEdit_notes.setText("sam_notes")

    expected_item = BPlan(
        "energy_scan",
        energies=energies,
        exposure=exposures,
        E0=11500.8,
        detectors=["vortex_me4", "I0"],
        md={
            "sample": "sam",
            "is_standard": False,
            "notes": "sam_notes",
        },
    )

    def check_item(item):
        item_dict = item.to_dict()["kwargs"]
        expected_dict = expected_item.to_dict()["kwargs"]

        try:
            # Check whether time is calculated correctly for a single scan
            assert display.ui.label_hour_scan.text() == "0"
            assert display.ui.label_min_scan.text() == "0"
            assert display.ui.label_sec_scan.text() == "27.8"

            # Check whether time is calculated correctly including the repeated scan
            assert display.ui.label_hour_total.text() == "0"
            assert display.ui.label_min_total.text() == "1"
            assert display.ui.label_sec_total.text() == "23.4"

            # Check energies & exposures within 3 decimals
            np.testing.assert_array_almost_equal(
                item_dict["energies"], expected_dict["energies"], decimal=2
            )
            np.testing.assert_array_almost_equal(
                item_dict["exposure"], expected_dict["exposure"], decimal=2
            )

            # Now check the rest of the dictionary, excluding the numpy array keys
            item_dict.pop("energies")
            item_dict.pop("exposure")
            expected_dict.pop("energies")
            expected_dict.pop("exposure")

            # Check if the remaining dictionary items are equal
            assert item_dict == expected_dict, "Non-array items do not match."

        except AssertionError as e:
            print(e)
            return False
        return True

    # Click the run button and see if the plan is queued
    display.ui.run_button.setEnabled(True)
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from ophyd_async.core import Device, StandardReadable

from .motor import Motor
from .transform import TransformRecord


class Table(StandardReadable, Device):
    """A dynamic Ophyd device for an optical table.

    This device will have the components needed for the configuration
    of motors provided. For example, if *horizontal_motor* is
    provided, then the device will have a ``horizontal`` attribute
    that is the corresponding horizontal motor.

    There are several dynamic parameters when creating an object that
    control the structure of the resulting device. These are described
    individually below and are assumed relative to *prefix*. If
    *upstream_motor* and *downstream_motor* are not empty, then
    *vertical* and *pitch* will point to a pseudo motor who's PV is
    determined by *pseudo_motors* (the value of *vertical_motor* will
    be ignored in this case). For example:

    .. code:: python

        tbl = Table(
            pseudo_motors="255idcVME:table_us:",
            transforms="255idcVME:table_us_trans:",
            upstream_motor="255idcVME:m3",
            downstream_motor="255idcVME:m4",
            name="my_table",
        )

    will produce a table with the following components:

    - *upstream*: "255idcVME:m3"
    - *downstream*: "255idcVME:m4"
    - *vertical*: "255idcVME:table_us:height"
    - *pitch*: "255idcVME:table_us:pitch"
    - *drive*: transform "255idcVME:table_us_trans:Drive"
    - *readback*: transform "255idcVME:table_us_trans:Readback"

    Parameters
    ==========
    vertical_motor
      The suffix for the real motor controlling the height of the
      table (e.g. "255idVME:m1"). Ignored if *upstream_motor* and
      *downstream_motor* are provided.
    horizontal_motor
      The suffix for the real motor controlling the lateral movement
      of the table (e.g. "255idVME:m2").
    upstream_motor
      The suffix for the real motor controlling the height of the
      upstream leg of the table (e.g. "255idVME:m3").
    downstream_motor
      The suffix for the real motor controlling the height of the
      downstream leg of the table (e.g. "255idVME:m4").
    pseudo_motors
      The prefix for the pseudo motors controlling the orientation of
      the table. E.g. "255id:table_ds:". Creates the components: *height*
      and *pitch*.
    transforms
      The prefix for the pseudo motor transform records
      (e.g. "255idVME:table_ds_trans:"). Creates the components:
      *vertical_drive_transform" and "vertical_readback_transform".

    """

    _ophyd_labels_ = {"tables"}
    # These are the possible components that could be present
    vertical: Device
    horizontal: Device
    upstream: Device
    downstream: Device
    pitch: Device
    horizontal_drive_transform: Device
    horizontal_readback_transform: Device

    def __init__(
        self,
        *,
        vertical_prefix: str = "",
        horizontal_prefix: str = "",
        upstream_prefix: str = "",
        downstream_prefix: str = "",
        pseudo_motor_prefix: str = "",
        transform_prefix: str = "",
        name="",
    ):
        has_pseudo_motors = bool(upstream_prefix) and bool(downstream_prefix)
        with self.add_children_as_readables():
            # See if we have direct control over the vertical/horizontal positions
            if vertical_prefix != "":
                self.vertical = Motor(vertical_prefix, name="vertical")
            if horizontal_prefix != "":
                self.horizontal = Motor(horizontal_prefix, name="horizontal")
            if bool(upstream_prefix):
                self.upstream = Motor(upstream_prefix, name="upstream")
            if bool(downstream_prefix):
                self.downstream = Motor(downstream_prefix, name="downstream")

            # Check if we need to add the pseudo motors and tranforms
            if has_pseudo_motors:
                self.vertical = Motor(f"{pseudo_motor_prefix}height", name="vertical")
                self.pitch = Motor(f"{pseudo_motor_prefix}pitch", name="pitch")
                self.vertical_drive_transform = TransformRecord(
                    f"{transform_prefix}Drive", name="vertical_drive_transform"
                )
                self.vertical_readback_transform = TransformRecord(
                    f"{transform_prefix}Readback", name="vertical_readback_transform"
                )
        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import PVGroup, SubGroup, ioc_arg_parser, pvproperty, run


class PreampsGroup(PVGroup):
    class PreampGroup(PVGroup):
        """An IOC with an SR570 pre-amplifier.

        E.g.

        - 25idc:SR01:IpreSlit:sens_num.VAL"
        - 25idc:SR01:IpreSlit:sens_unit.VAL

        """

        sens_num = pvproperty(
            value="5",
            enum_strings=["1", "2", "5", "10", "20", "50", "100", "200", "500"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        sens_unit = pvproperty(
            value="nA/V",
            enum_strings=["pA/V", "nA/V", "uA/V", "mA/V"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        offset_num = pvproperty(
            value="5",
            enum_strings=["1", "2", "5", "10", "20", "50", "100", "200", "500"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        offset_unit = pvproperty(
            value="nA",
            enum_strings=["pA", "nA", "uA", "mA"],
            record="mbbi",
            dtype=ChannelType.ENUM,
        )
        set_all = pvproperty(value=1, name="init.PROC", doc="")

    preamp1 = SubGroup(PreampGroup, prefix="SR01:")
    preamp2 = SubGroup(PreampGroup, prefix="SR02:")
    preamp3 = SubGroup(PreampGroup, prefix="SR03:")
    preamp4 = SubGroup(PreampGroup, prefix="SR04:")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idc:", desc="haven.tests.ioc_preamp test IOC"
    )
    ioc = PreampsGroup(**ioc_options)
    run(ioc.pvdb, **run_options)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.devices.beamline_manager import BeamlineManager

"""
    glados:alive.VAL
    glados:alive.RHOST
    glados:alive.RADDR
    glados:alive.RPORT
    glados:alive.RRSTS
    glados:alive.AHOST
    glados:alive.AADDR
    glados:alive.APORT
    glados:alive.ARSTS
    glados:alive.HRTBT
    glados:alive.HPRD
    glados:alive.IOCNM
    glados:alive.HMAG
    glados:alive.MSG
    glados:alive.IPORT
    glados:alive.IPSTS
    glados:alive.ITRIG
    glados:alive.ISUP
    glados:alive.VER
    glados:alive.EVD1
    glados:alive.EVD2
    glados:alive.EVD3
    glados:alive.EVD4
    glados:alive.EVD5
    glados:alive.EVD6
    glados:alive.EVD7
    glados:alive.EVD8
    glados:alive.EVD9
    glados:alive.EVD10
    glados:alive.EVD11
    glados:alive.EVD12
    glados:alive.EVD13
    glados:alive.EVD14
    glados:alive.EVD15
    glados:alive.EVD16
    glados:alive.EV1
    glados:alive.EV2
    glados:alive.EV3
    glados:alive.EV4
    glados:alive.EV5
    glados:alive.EV6
    glados:alive.EV7
    glados:alive.EV8
    glados:alive.EV9
    glados:alive.EV10
    glados:alive.EV11
    glados:alive.EV12
    glados:alive.EV13
    glados:alive.EV14
    glados:alive.EV15
    glados:alive.EV16
    glados:bss:proposal:beamline
    glados:bss:proposal:endDate
    glados:bss:proposal:endTimestamp
    glados:bss:proposal:mailInFlag
    glados:bss:proposal:id
    glados:bss:proposal:proprietaryFlag
    glados:bss:proposal:raw
    glados:bss:proposal:startDate
    glados:bss:proposal:startTimestamp
    glados:bss:proposal:submittedDate
    glados:bss:proposal:submittedTimestamp
    glados:bss:proposal:title
    glados:bss:proposal:userPIs
    glados:bss:proposal:userBadges
    glados:bss:proposal:users
    glados:bss:proposal:users_in_pvs
    glados:bss:proposal:users_total
    glados:bss:proposal:user1:badgeNumber
    glados:bss:proposal:user1:email
    glados:bss:proposal:user1:firstName
    glados:bss:proposal:user1:lastName
    glados:bss:proposal:user1:piFlag
    glados:bss:proposal:user1:institution
    glados:bss:proposal:user2:badgeNumber
    glados:bss:proposal:user2:email
    glados:bss:proposal:user2:firstName
    glados:bss:proposal:user2:lastName
    glados:bss:proposal:user2:piFlag
    glados:bss:proposal:user2:institution
    glados:bss:proposal:user3:badgeNumber
    glados:bss:proposal:user3:email
    glados:bss:proposal:user3:firstName
    glados:bss:proposal:user3:lastName
    glados:bss:proposal:user3:piFlag
    glados:bss:proposal:user3:institution
    glados:bss:proposal:user4:badgeNumber
    glados:bss:proposal:user4:email
    glados:bss:proposal:user4:firstName
    glados:bss:proposal:user4:lastName
    glados:bss:proposal:user4:piFlag
    glados:bss:proposal:user4:institution
    glados:bss:proposal:user5:badgeNumber
    glados:bss:proposal:user5:email
    glados:bss:proposal:user5:firstName
    glados:bss:proposal:user5:lastName
    glados:bss:proposal:user5:piFlag
    glados:bss:proposal:user5:institution
    glados:bss:proposal:user6:badgeNumber
    glados:bss:proposal:user6:email
    glados:bss:proposal:user6:firstName
    glados:bss:proposal:user6:lastName
    glados:bss:proposal:user6:piFlag
    glados:bss:proposal:user6:institution
    glados:bss:proposal:user7:badgeNumber
    glados:bss:proposal:user7:email
    glados:bss:proposal:user7:firstName
    glados:bss:proposal:user7:lastName
    glados:bss:proposal:user7:piFlag
    glados:bss:proposal:user7:institution
    glados:bss:proposal:user8:badgeNumber
    glados:bss:proposal:user8:email
    glados:bss:proposal:user8:firstName
    glados:bss:proposal:user8:lastName
    glados:bss:proposal:user8:piFlag
    glados:bss:proposal:user8:institution
    glados:bss:proposal:user9:badgeNumber
    glados:bss:proposal:user9:email
    glados:bss:proposal:user9:firstName
    glados:bss:proposal:user9:lastName
    glados:bss:proposal:user9:piFlag
    glados:bss:proposal:user9:institution
    glados:bss:esaf:cycle
    glados:bss:esaf:description
    glados:bss:esaf:endDate
    glados:bss:esaf:endTimestamp
    glados:bss:esaf:id
    glados:bss:esaf:raw
    glados:bss:esaf:status
    glados:bss:esaf:sector
    glados:bss:esaf:startDate
    glados:bss:esaf:startTimestamp
    glados:bss:esaf:title
    glados:bss:esaf:userBadges
    glados:bss:esaf:userPIs
    glados:bss:esaf:users
    glados:bss:esaf:users_in_pvs
    glados:bss:esaf:users_total
    glados:bss:esaf:user1:badgeNumber
    glados:bss:esaf:user1:email
    glados:bss:esaf:user1:firstName
    glados:bss:esaf:user1:lastName
    glados:bss:esaf:user2:badgeNumber
    glados:bss:esaf:user2:email
    glados:bss:esaf:user2:firstName
    glados:bss:esaf:user2:lastName
    glados:bss:esaf:user3:badgeNumber
    glados:bss:esaf:user3:email
    glados:bss:esaf:user3:firstName
    glados:bss:esaf:user3:lastName
    glados:bss:esaf:user4:badgeNumber
    glados:bss:esaf:user4:email
    glados:bss:esaf:user4:firstName
    glados:bss:esaf:user4:lastName
    glados:bss:esaf:user5:badgeNumber
    glados:bss:esaf:user5:email
    glados:bss:esaf:user5:firstName
    glados:bss:esaf:user5:lastName
    glados:bss:esaf:user6:badgeNumber
    glados:bss:esaf:user6:email
    glados:bss:esaf:user6:firstName
    glados:bss:esaf:user6:lastName
    glados:bss:esaf:user7:badgeNumber
    glados:bss:esaf:user7:email
    glados:bss:esaf:user7:firstName
    glados:bss:esaf:user7:lastName
    glados:bss:esaf:user8:badgeNumber
    glados:bss:esaf:user8:email
    glados:bss:esaf:user8:firstName
    glados:bss:esaf:user8:lastName
    glados:bss:esaf:user9:badgeNumber
    glados:bss:esaf:user9:email
    glados:bss:esaf:user9:firstName
    glados:bss:esaf:user9:lastName
    glados:bss:status
    glados:bss:ioc_host
    glados:bss:ioc_user
    glados:ioc25ida:startable
    glados:ioc25ida:start
    glados:ioc25ida:stoppable
    glados:ioc25ida:stop
    glados:ioc25ida:restart
    glados:ioc25ida:status
    glados:ioc25ida:console
    glados:ioc25idaBPM:startable
    glados:ioc25idaBPM:start
    glados:ioc25idaBPM:stoppable
    glados:ioc25idaBPM:stop
    glados:ioc25idaBPM:restart
    glados:ioc25idaBPM:status
    glados:ioc25idaBPM:console
    glados:ioc25idVac:startable
    glados:ioc25idVac:start
    glados:ioc25idVac:stoppable
    glados:ioc25idVac:stop
    glados:ioc25idVac:restart
    glados:ioc25idVac:status
    glados:ioc25idVac:console
    glados:ioc25idbUP:startable
    glados:ioc25idbUP:start
    glados:ioc25idbUP:stoppable
    glados:ioc25idbUP:stop
    glados:ioc25idbUP:restart
    glados:ioc25idbUP:status
    glados:ioc25idbUP:console
    glados:ioc25idc:startable
    glados:ioc25idc:start
    glados:ioc25idc:stoppable
    glados:ioc25idc:stop
    glados:ioc25idc:restart
    glados:ioc25idc:status
    glados:ioc25idc:console
    glados:ioc25idgigeA:startable
    glados:ioc25idgigeA:start
    glados:ioc25idgigeA:stoppable
    glados:ioc25idgigeA:stop
    glados:ioc25idgigeA:restart
    glados:ioc25idgigeA:status
    glados:ioc25idgigeA:console
    glados:ioc25idgigeB:startable
    glados:ioc25idgigeB:start
    glados:ioc25idgigeB:stoppable
    glados:ioc25idgigeB:stop
    glados:ioc25idgigeB:restart
    glados:ioc25idgigeB:status
    glados:ioc25idgigeB:console
    glados:ioc25idgigeC:startable
    glados:ioc25idgigeC:start
    glados:ioc25idgigeC:stoppable
    glados:ioc25idgigeC:stop
    glados:ioc25idgigeC:restart
    glados:ioc25idgigeC:status
    glados:ioc25idgigeC:console
    glados:ioc25idgigeD:startable
    glados:ioc25idgigeD:start
    glados:ioc25idgigeD:stoppable
    glados:ioc25idgigeD:stop
    glados:ioc25idgigeD:restart
    glados:ioc25idgigeD:status
    glados:ioc25idgigeD:console
    glados:ioc25idgigeE:startable
    glados:ioc25idgigeE:start
    glados:ioc25idgigeE:stoppable
    glados:ioc25idgigeE:stop
    glados:ioc25idgigeE:restart
    glados:ioc25idgigeE:status
    glados:ioc25idgigeE:console
    glados:ioc25idSimDet:startable
    glados:ioc25idSimDet:start
    glados:ioc25idSimDet:stoppable
    glados:ioc25idSimDet:stop
    glados:ioc25idSimDet:restart
    glados:ioc25idSimDet:status
    glados:ioc25idSimDet:console
    glados:local_storage:file_system
    glados:local_storage:sub_directory
    glados:local_storage:full_path
    glados:local_storage:exists
    glados:local_storage:create
    glados:local_storage_startup
    glados:autosave_helper::__autosave_hook
"""


def test_iocs():
    """Check that the managed IOCs are available."""
    iocs = {
        "ioc255idc": "ioc255idc:",
    }
    manager = BeamlineManager("255id:glados:", name="glados", iocs=iocs)
    assert hasattr(manager, "iocs")
    assert manager.iocs.ioc255idc.prefix == "255id:glados:ioc255idc:"



import asyncio

import pytest
from ophyd_async.core import set_mock_value

from haven.devices.energy_positioner import EnergyPositioner
from haven.devices.xray_source import BusyStatus


@pytest.fixture()
async def positioner():
    positioner = EnergyPositioner(
        name="energy",
        monochromator_prefix="255idMono:",
        undulator_prefix="S255ID:",
    )
    await positioner.connect(mock=True)
    return positioner


async def test_set_energy(positioner):
    # Set up dependent values
    set_mock_value(positioner.monochromator.id_offset, 150)
    # Change the energy
    status = positioner.set(10000, timeout=3)
    # Trick the Undulator into being done
    set_mock_value(positioner.undulator.energy.done, BusyStatus.BUSY)
    await asyncio.sleep(0.01)  # Let the event loop run
    set_mock_value(positioner.undulator.energy.done, BusyStatus.DONE)
    await status
    # Check that all the sub-components were set properly
    assert await positioner.monochromator.energy.user_setpoint.get_value() == 10000
    assert await positioner.undulator.energy.setpoint.get_value() == 10.150


async def test_real_to_pseudo_positioner(positioner):
    set_mock_value(positioner.monochromator.energy.user_readback, 5000.0)
    # Check that the pseudo single is updated
    reading = await positioner.read()
    assert reading["energy"]["value"] == 5000.0


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
